"""
RT-DETR v1 (Real-Time DEtection TRansformer), adapted from
https://github.com/lyuwenyu/RT-DETR/tree/main/rtdetr_pytorch

Paper "DETRs Beat YOLOs on Real-time Object Detection", https://arxiv.org/abs/2304.08069

Note: The paper describes uncertainty-minimal query selection, but the upstream repo
selects top-K queries by classification confidence. We follow the repo behavior here.
"""

# Reference license: Apache-2.0

import copy
import math
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import boxes as box_ops

from birder.common import training_utils
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import pos_embedding_sin_cos_2d
from birder.net.detection.base import DetectionBaseNet
from birder.net.detection.deformable_detr import DeformableTransformerDecoderLayer
from birder.net.detection.deformable_detr import HungarianMatcher
from birder.net.detection.deformable_detr import inverse_sigmoid
from birder.net.repvgg import RepVggBlock


def _get_clones(module: nn.Module, N: int) -> nn.ModuleList:
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


@torch.jit.unused  # type: ignore[untyped-decorator]
def get_contrastive_denoising_training_group(  # pylint: disable=too-many-locals
    targets: list[dict[str, torch.Tensor]],
    num_classes: int,
    num_queries: int,
    class_embed: nn.Embedding,
    num_denoising_queries: int,
    label_noise_ratio: float,
    box_noise_scale: float,
) -> tuple[Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor], Optional[dict[str, Any]]]:
    num_ground_truths = [len(t["labels"]) for t in targets]
    device = targets[0]["labels"].device

    max_gt_num = max(num_ground_truths)
    if max_gt_num == 0:
        return (None, None, None, None)

    num_groups = num_denoising_queries // max_gt_num
    num_groups = 1 if num_groups == 0 else num_groups

    # Pad ground truth to max_num of a batch
    batch_size = len(num_ground_truths)

    input_query_class = torch.full([batch_size, max_gt_num], num_classes, dtype=torch.long, device=device)
    input_query_bbox = torch.zeros([batch_size, max_gt_num, 4], dtype=targets[0]["boxes"].dtype, device=device)
    pad_gt_mask = torch.zeros([batch_size, max_gt_num], dtype=torch.bool, device=device)

    for i in range(batch_size):
        num_gt = num_ground_truths[i]
        if num_gt > 0:
            input_query_class[i, :num_gt] = targets[i]["labels"].long()
            input_query_bbox[i, :num_gt] = targets[i]["boxes"]
            pad_gt_mask[i, :num_gt] = True

    input_query_class = input_query_class.tile([1, 2 * num_groups])
    input_query_bbox = input_query_bbox.tile([1, 2 * num_groups, 1])
    pad_gt_mask = pad_gt_mask.tile([1, 2 * num_groups])

    # Positive and negative mask
    negative_gt_mask = torch.zeros([batch_size, max_gt_num * 2, 1], device=device)
    negative_gt_mask[:, max_gt_num:] = 1
    negative_gt_mask = negative_gt_mask.tile([1, num_groups, 1])
    positive_gt_mask = 1 - negative_gt_mask

    # Contrastive denoising training positive index
    positive_gt_mask_squeezed = positive_gt_mask.squeeze(-1) * pad_gt_mask
    denoise_positive_idx = torch.nonzero(positive_gt_mask_squeezed)[:, 1]
    denoise_positive_idx = torch.split(denoise_positive_idx, [n * num_groups for n in num_ground_truths])

    # Total denoising queries
    total_denoising_queries = int(max_gt_num * 2 * num_groups)

    # Apply label noise
    if label_noise_ratio > 0:
        mask = torch.rand_like(input_query_class, dtype=torch.float) < (label_noise_ratio * 0.5)
        new_label = torch.randint(0, num_classes, input_query_class.shape, dtype=torch.long, device=device)
        input_query_class = torch.where(mask & pad_gt_mask, new_label, input_query_class)

    # Apply box noise
    if box_noise_scale > 0:
        known_bbox = box_ops.box_convert(input_query_bbox, in_fmt="cxcywh", out_fmt="xyxy")
        diff = torch.tile(input_query_bbox[..., 2:] * 0.5, [1, 1, 2]) * box_noise_scale
        rand_sign = torch.randint_like(input_query_bbox, 0, 2) * 2.0 - 1.0
        rand_part = torch.rand_like(input_query_bbox)
        rand_part = (rand_part + 1.0) * negative_gt_mask + rand_part * (1 - negative_gt_mask)
        rand_part *= rand_sign
        known_bbox = known_bbox + rand_part * diff
        known_bbox.clip_(min=0.0, max=1.0)
        input_query_bbox = box_ops.box_convert(known_bbox, in_fmt="xyxy", out_fmt="cxcywh")
        input_query_bbox = inverse_sigmoid(input_query_bbox)

    # Embed class labels
    input_query_class = class_embed(input_query_class)

    # Create attention mask (True = masked)
    target_size = total_denoising_queries + num_queries
    attn_mask = torch.zeros([target_size, target_size], dtype=torch.bool, device=device)
    attn_mask[total_denoising_queries:, :total_denoising_queries] = True
    for i in range(num_groups):
        idx_block_start = max_gt_num * 2 * i
        idx_block_end = max_gt_num * 2 * (i + 1)
        attn_mask[idx_block_start:idx_block_end, :idx_block_start] = True
        attn_mask[idx_block_start:idx_block_end, idx_block_end:total_denoising_queries] = True

    denoising_meta_values = {
        "dn_positive_idx": denoise_positive_idx,
        "dn_num_group": num_groups,
        "dn_num_split": [total_denoising_queries, num_queries],
    }

    return (input_query_class, input_query_bbox, attn_mask, denoising_meta_values)


def varifocal_loss(
    pred_logits: torch.Tensor,
    target_score: torch.Tensor,
    target_onehot: torch.Tensor,
    alpha: float = 0.75,
    gamma: float = 2.0,
) -> torch.Tensor:
    """
    VariFocal Loss - uses IoU as soft target for positives, focal-like weighting for negatives
    """

    pred_score = pred_logits.sigmoid().detach()
    weight = alpha * pred_score.pow(gamma) * (1 - target_onehot) + target_score
    loss = F.binary_cross_entropy_with_logits(pred_logits, target_score, weight=weight, reduction="none")

    return loss


class CSPRepLayer(nn.Module):
    """
    Cross Stage Partial layer with RepVGG blocks
    """

    def __init__(self, in_channels: int, out_channels: int, num_blocks: int, expansion: float) -> None:
        super().__init__()
        hidden_channels = int(out_channels * expansion)

        self.conv1 = Conv2dNormActivation(
            in_channels,
            hidden_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            activation_layer=nn.SiLU,
        )
        self.conv2 = Conv2dNormActivation(
            in_channels,
            hidden_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            activation_layer=nn.SiLU,
        )
        self.bottlenecks = nn.Sequential(
            *[
                RepVggBlock(
                    in_channels=hidden_channels,
                    out_channels=hidden_channels,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    groups=1,
                    use_se=False,
                    reparameterized=False,
                    act_layer=nn.SiLU,
                    skip_identity=True,
                )
                for _ in range(num_blocks)
            ]
        )

        if hidden_channels != out_channels:
            self.conv3: nn.Module = Conv2dNormActivation(
                hidden_channels,
                out_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                activation_layer=nn.SiLU,
            )
        else:
            self.conv3 = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x1 = self.bottlenecks(self.conv1(x))
        x2 = self.conv2(x)
        return self.conv3(x1 + x2)


class MultiheadAttention(nn.Module):
    def __init__(self, d_model: int, num_heads: int, attn_drop: float = 0.0, proj_drop: float = 0.0) -> None:
        super().__init__()
        assert d_model % num_heads == 0, "d_model should be divisible by num_heads"

        self.num_heads = num_heads
        self.head_dim = d_model // num_heads
        self.scale = self.head_dim**-0.5

        self.q_proj = nn.Linear(d_model, d_model)
        self.k_proj = nn.Linear(d_model, d_model)
        self.v_proj = nn.Linear(d_model, d_model)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(d_model, d_model)
        self.proj_drop = nn.Dropout(proj_drop)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.xavier_uniform_(self.q_proj.weight)
        nn.init.xavier_uniform_(self.k_proj.weight)
        nn.init.xavier_uniform_(self.v_proj.weight)
        nn.init.xavier_uniform_(self.proj.weight)
        if self.q_proj.bias is not None:
            nn.init.zeros_(self.q_proj.bias)
            nn.init.zeros_(self.k_proj.bias)
            nn.init.zeros_(self.v_proj.bias)
            nn.init.zeros_(self.proj.bias)

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        key_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        B, l_q, C = query.shape
        q = self.q_proj(query).reshape(B, l_q, self.num_heads, self.head_dim).transpose(1, 2)
        k = self.k_proj(key).reshape(B, key.size(1), self.num_heads, self.head_dim).transpose(1, 2)
        v = self.v_proj(value).reshape(B, value.size(1), self.num_heads, self.head_dim).transpose(1, 2)

        if key_padding_mask is not None:
            # key_padding_mask is expected to be boolean (True = masked)
            # SDPA expects True = attend, so we invert
            attn_mask = ~key_padding_mask[:, None, None, :]
        else:
            attn_mask = None

        attn = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, attn_mask=attn_mask, dropout_p=self.attn_drop.p if self.training else 0.0, scale=self.scale
        )

        attn = attn.transpose(1, 2).reshape(B, l_q, C)
        x = self.proj(attn)
        x = self.proj_drop(x)

        return x


class TransformerEncoderLayer(nn.Module):
    def __init__(self, d_model: int, num_heads: int, dim_feedforward: int, dropout: float) -> None:
        super().__init__()
        self.self_attn = MultiheadAttention(d_model, num_heads, attn_drop=dropout)

        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.activation = nn.GELU()

    def forward(
        self, src: torch.Tensor, pos: torch.Tensor, key_padding_mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        qk = src + pos
        src2 = self.self_attn(qk, qk, value=src, key_padding_mask=key_padding_mask)
        src = src + self.dropout1(src2)
        src = self.norm1(src)

        src2 = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = src + self.dropout2(src2)
        src = self.norm2(src)

        return src


class AIFI(nn.Module):
    """
    Attention-based Intra-scale Feature Interaction.
    Applies transformer encoder to a single feature scale.
    """

    def __init__(self, d_model: int, num_heads: int, dim_feedforward: int, dropout: float, num_layers: int) -> None:
        super().__init__()
        self.d_model = d_model
        encoder_layer = TransformerEncoderLayer(d_model, num_heads, dim_feedforward, dropout)
        self.layers = _get_clones(encoder_layer, num_layers)
        self.use_cache = True
        self._pos_cache: dict[str, torch.Tensor] = {}

    def set_cache_enabled(self, enabled: bool) -> None:
        self.use_cache = enabled
        if enabled is False:
            self.clear_cache()

    def clear_cache(self) -> None:
        self._pos_cache.clear()

    def forward(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        B, C, H, W = x.size()
        x = x.flatten(2).permute(0, 2, 1)

        use_cache = self.use_cache is True and torch.jit.is_tracing() is False and torch.jit.is_scripting() is False
        if use_cache is True:
            cache_key = f"{H}x{W}x{C}_{x.device}_{x.dtype}"
            pos = self._pos_cache.get(cache_key)
            if pos is None:
                pos = pos_embedding_sin_cos_2d(H, W, C, num_special_tokens=0, device=x.device).unsqueeze(0)
                self._pos_cache[cache_key] = pos
        else:
            pos = pos_embedding_sin_cos_2d(H, W, C, num_special_tokens=0, device=x.device).unsqueeze(0)

        key_padding_mask = mask.flatten(1) if mask is not None else None
        for layer in self.layers:
            x = layer(x, pos, key_padding_mask=key_padding_mask)

        x = x.permute(0, 2, 1).reshape(B, C, H, W)

        return x


class HybridEncoder(nn.Module):
    """
    RT-DETR Hybrid Encoder: AIFI + CCFM (CNN-based Cross-scale Feature-fusion Module).

    Architecture:
    1. Input projection for each scale
    2. AIFI (transformer) on the highest resolution feature
    3. Top-down FPN path with CSPRepLayers
    4. Bottom-up PAN path with CSPRepLayers
    """

    def __init__(
        self,
        in_channels: list[int],
        hidden_dim: int,
        num_encoder_layers: int,
        dim_feedforward: int,
        dropout: float,
        num_heads: int,
        expansion: float,
        depth_multiplier: float,
    ) -> None:
        super().__init__()
        self.num_levels = len(in_channels)

        # Input projection layers
        self.input_proj = nn.ModuleList()
        for ch in in_channels:
            self.input_proj.append(
                nn.Sequential(
                    nn.Conv2d(ch, hidden_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
                    nn.BatchNorm2d(hidden_dim),
                )
            )

        # AIFI encoder
        self.encoder = AIFI(
            hidden_dim,
            num_heads=num_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            num_layers=num_encoder_layers,
        )

        # CCFM: Top-down FPN path
        num_blocks = max(round(3 * depth_multiplier), 1)
        self.lateral_convs = nn.ModuleList()
        self.fpn_blocks = nn.ModuleList()
        for _ in range(self.num_levels - 1, 0, -1):
            self.lateral_convs.append(
                Conv2dNormActivation(
                    hidden_dim,
                    hidden_dim,
                    kernel_size=(1, 1),
                    stride=(1, 1),
                    padding=(0, 0),
                    activation_layer=nn.SiLU,
                )
            )
            self.fpn_blocks.append(CSPRepLayer(hidden_dim * 2, hidden_dim, num_blocks=num_blocks, expansion=expansion))

        # CCFM: Bottom-up PAN path
        self.downsample_convs = nn.ModuleList()
        self.pan_blocks = nn.ModuleList()
        for _ in range(self.num_levels - 1):
            self.downsample_convs.append(
                Conv2dNormActivation(
                    hidden_dim,
                    hidden_dim,
                    kernel_size=(3, 3),
                    stride=(2, 2),
                    padding=(1, 1),
                    activation_layer=nn.SiLU,
                )
            )
            self.pan_blocks.append(CSPRepLayer(hidden_dim * 2, hidden_dim, num_blocks=num_blocks, expansion=expansion))

    def forward(self, feats: list[torch.Tensor], masks: Optional[list[torch.Tensor]] = None) -> list[torch.Tensor]:
        # Project all features to hidden_dim
        proj_feats = []
        for idx, proj in enumerate(self.input_proj):
            proj_feats.append(proj(feats[idx]))

        # Apply AIFI to highest scale feature
        encoder_mask = masks[-1] if masks is not None else None
        proj_feats[-1] = self.encoder(proj_feats[-1], mask=encoder_mask)

        # Top-down FPN path
        inner_outs = [proj_feats[-1]]
        for idx, (lateral_conv, fpn_block) in enumerate(zip(self.lateral_convs, self.fpn_blocks)):
            feat_high = inner_outs[0]
            feat_low = proj_feats[self.num_levels - 2 - idx]

            feat_high = lateral_conv(feat_high)
            inner_outs[0] = feat_high
            feat_high = F.interpolate(feat_high, size=feat_low.shape[-2:], mode="nearest")

            inner_out = fpn_block(torch.concat([feat_high, feat_low], dim=1))
            inner_outs.insert(0, inner_out)

        # Bottom-up PAN path
        outs = [inner_outs[0]]
        for idx, (downsample_conv, pan_block) in enumerate(zip(self.downsample_convs, self.pan_blocks)):
            feat_low = outs[-1]
            feat_high = inner_outs[idx + 1]
            feat_low = downsample_conv(feat_low)
            out = pan_block(torch.concat([feat_low, feat_high], dim=1))
            outs.append(out)

        return outs

    def set_cache_enabled(self, enabled: bool) -> None:
        self.encoder.set_cache_enabled(enabled)

    def clear_cache(self) -> None:
        self.encoder.clear_cache()


# pylint: disable=invalid-name
class RT_DETRDecoder(nn.Module):
    """
    RT-DETR Decoder with uncertainty-minimal query selection.

    Key difference from Deformable DETR: queries are selected from encoder output
    based on classification confidence, rather than using learned query embeddings.
    """

    def __init__(
        self,
        hidden_dim: int,
        num_classes: int,
        num_queries: int,
        num_decoder_layers: int,
        num_levels: int,
        num_heads: int,
        dim_feedforward: int,
        dropout: float,
        num_decoder_points: int,
    ) -> None:
        super().__init__()
        self.hidden_dim = hidden_dim
        self.num_queries = num_queries

        self.enc_output = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.LayerNorm(hidden_dim),
        )
        self.enc_score_head = nn.Linear(hidden_dim, num_classes)
        self.enc_bbox_head = MLP(hidden_dim, [hidden_dim, hidden_dim, 4], activation_layer=nn.ReLU)

        decoder_layer = DeformableTransformerDecoderLayer(
            hidden_dim, dim_feedforward, dropout, num_levels, num_heads, num_decoder_points
        )
        self.layers = _get_clones(decoder_layer, num_decoder_layers)

        self.query_pos_head = MLP(4, [2 * hidden_dim, hidden_dim], activation_layer=nn.ReLU)
        self.class_embed = nn.ModuleList([nn.Linear(hidden_dim, num_classes) for _ in range(num_decoder_layers)])
        self.bbox_embed = nn.ModuleList(
            [MLP(hidden_dim, [hidden_dim, hidden_dim, 4], activation_layer=nn.ReLU) for _ in range(num_decoder_layers)]
        )
        self.use_cache = True
        self._anchor_cache: dict[str, tuple[torch.Tensor, torch.Tensor]] = {}

        # Weights initialization
        prior_prob = 0.01
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        nn.init.xavier_uniform_(self.enc_output[0].weight)
        nn.init.xavier_uniform_(self.enc_score_head.weight)
        nn.init.constant_(self.enc_score_head.bias, bias_value)
        nn.init.zeros_(self.enc_bbox_head[-2].weight)
        nn.init.zeros_(self.enc_bbox_head[-2].bias)
        for class_embed in self.class_embed:
            nn.init.constant_(class_embed.bias, bias_value)

        for bbox_embed in self.bbox_embed:
            nn.init.zeros_(bbox_embed[-2].weight)
            nn.init.zeros_(bbox_embed[-2].bias)

    def set_cache_enabled(self, enabled: bool) -> None:
        self.use_cache = enabled
        if enabled is False:
            self.clear_cache()

    def clear_cache(self) -> None:
        self._anchor_cache.clear()

    def _generate_anchors(
        self,
        spatial_shapes: list[list[int]],
        grid_size: float = 0.05,
        device: torch.device = torch.device("cpu"),
        dtype: torch.dtype = torch.float32,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        cache_key: Optional[str] = None
        use_cache = self.use_cache is True and torch.jit.is_tracing() is False and torch.jit.is_scripting() is False
        if use_cache is True:
            spatial_key = ",".join(f"{int(h)}x{int(w)}" for h, w in spatial_shapes)
            cache_key = f"{spatial_key}_{grid_size}_{device}_{dtype}"
            cached = self._anchor_cache.get(cache_key)
            if cached is not None:
                return cached

        anchors = []
        for lvl, (h, w) in enumerate(spatial_shapes):
            grid_y, grid_x = torch.meshgrid(
                torch.arange(h, dtype=dtype, device=device),
                torch.arange(w, dtype=dtype, device=device),
                indexing="ij",
            )
            grid_xy = torch.stack([grid_x, grid_y], dim=-1)
            valid_wh = torch.tensor([w, h], dtype=dtype, device=device)
            grid_xy = (grid_xy.unsqueeze(0) + 0.5) / valid_wh
            wh = torch.ones_like(grid_xy) * grid_size * (2.0**lvl)
            anchors.append(torch.concat([grid_xy, wh], dim=-1).reshape(-1, h * w, 4))

        anchors = torch.concat(anchors, dim=1)
        eps = 0.01
        valid_mask = ((anchors > eps) * (anchors < 1 - eps)).all(dim=-1, keepdim=True)
        anchors = torch.log(anchors / (1 - anchors))
        anchors = torch.where(valid_mask, anchors, torch.inf)

        if cache_key is not None:
            self._anchor_cache[cache_key] = (anchors, valid_mask)

        return (anchors, valid_mask)

    def _get_decoder_input(
        self,
        memory: torch.Tensor,
        spatial_shapes: list[list[int]],
        memory_padding_mask: Optional[torch.Tensor] = None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        anchors, valid_mask = self._generate_anchors(spatial_shapes, device=memory.device, dtype=memory.dtype)
        if memory_padding_mask is not None:
            valid_mask = valid_mask & ~memory_padding_mask.unsqueeze(-1)

        memory = valid_mask.to(memory.dtype) * memory
        output_memory = self.enc_output(memory)
        enc_outputs_class = self.enc_score_head(output_memory)
        if memory_padding_mask is not None:
            enc_outputs_class = enc_outputs_class.masked_fill(memory_padding_mask[..., None], float("-inf"))

        enc_outputs_coord_unact = self.enc_bbox_head(output_memory) + anchors

        # Select top-k queries based on classification confidence
        _, topk_ind = torch.topk(enc_outputs_class.max(dim=-1).values, self.num_queries, dim=1)

        # Gather reference points
        reference_points_unact = enc_outputs_coord_unact.gather(
            dim=1, index=topk_ind.unsqueeze(-1).expand(-1, -1, enc_outputs_coord_unact.shape[-1])
        )

        enc_topk_bboxes = reference_points_unact.sigmoid()

        # Gather encoder logits for loss computation
        enc_topk_logits = enc_outputs_class.gather(
            dim=1, index=topk_ind.unsqueeze(-1).expand(-1, -1, enc_outputs_class.shape[-1])
        )

        # Extract region features
        target = output_memory.gather(dim=1, index=topk_ind.unsqueeze(-1).expand(-1, -1, output_memory.shape[-1]))
        target = target.detach()

        return (target, reference_points_unact.detach(), enc_topk_bboxes, enc_topk_logits)

    def forward(  # pylint: disable=too-many-locals
        self,
        feats: list[torch.Tensor],
        spatial_shapes: list[list[int]],
        level_start_index: list[int],
        denoising_class: Optional[torch.Tensor] = None,
        denoising_bbox_unact: Optional[torch.Tensor] = None,
        attn_mask: Optional[torch.Tensor] = None,
        padding_mask: Optional[list[torch.Tensor]] = None,
        return_intermediates: bool = True,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        memory = []
        mask_flatten = []
        for idx, feat in enumerate(feats):
            feat = feat.flatten(2).permute(0, 2, 1)  # (B, H*W, C)
            memory.append(feat)
            if padding_mask is not None:
                mask_flatten.append(padding_mask[idx].flatten(1))

        memory = torch.concat(memory, dim=1)
        memory_padding_mask = torch.concat(mask_flatten, dim=1) if mask_flatten else None

        # Get decoder input (query selection)
        target, init_ref_points_unact, enc_topk_bboxes, enc_topk_logits = self._get_decoder_input(
            memory, spatial_shapes, memory_padding_mask
        )

        # Concatenate denoising queries if provided
        if denoising_class is not None and denoising_bbox_unact is not None:
            target = torch.concat([denoising_class, target], dim=1)
            init_ref_points_unact = torch.concat([denoising_bbox_unact, init_ref_points_unact], dim=1)

        # Prepare spatial shapes and level start index as tensors
        spatial_shapes_tensor = torch.tensor(spatial_shapes, dtype=torch.long, device=memory.device)
        level_start_index_tensor = torch.tensor(level_start_index, dtype=torch.long, device=memory.device)

        # Decoder forward
        bboxes_list: list[torch.Tensor] = []
        logits_list: list[torch.Tensor] = []
        reference_points = init_ref_points_unact.sigmoid()
        for decoder_layer, bbox_head, class_head in zip(self.layers, self.bbox_embed, self.class_embed):
            query_pos = self.query_pos_head(reference_points)
            reference_points_input = reference_points.unsqueeze(2).expand(-1, -1, len(spatial_shapes), -1)
            target = decoder_layer(
                target,
                query_pos,
                reference_points_input,
                memory,
                spatial_shapes_tensor,
                level_start_index_tensor,
                memory_padding_mask,
                attn_mask,
                src_shapes=spatial_shapes,
            )

            bbox_delta = bbox_head(target)
            new_reference_points = inverse_sigmoid(reference_points) + bbox_delta
            new_reference_points = new_reference_points.sigmoid()

            # Classification
            class_logits = class_head(target)

            if return_intermediates is True:
                bboxes_list.append(new_reference_points)
                logits_list.append(class_logits)

            # Update reference points for next layer
            reference_points = new_reference_points.detach()

        if return_intermediates is True:
            out_bboxes = torch.stack(bboxes_list)
            out_logits = torch.stack(logits_list)
        else:
            out_bboxes = new_reference_points
            out_logits = class_logits

        return (out_bboxes, out_logits, enc_topk_bboxes, enc_topk_logits)


# pylint: disable=invalid-name
class RT_DETR_v1(DetectionBaseNet):
    default_size = (640, 640)

    def __init__(
        self,
        num_classes: int,
        backbone: DetectorBackbone,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
        export_mode: bool = False,
    ) -> None:
        super().__init__(num_classes, backbone, config=config, size=size, export_mode=export_mode)
        assert self.config is not None, "must set config"

        self.reparameterized = False

        # Sigmoid based classification (no background class in predictions)
        self.num_classes = self.num_classes - 1

        hidden_dim = 256
        num_heads = 8
        dim_feedforward = 1024
        num_decoder_points = 4
        dropout: float = self.config.get("dropout", 0.0)
        num_encoder_layers: int = self.config.get("num_encoder_layers", 1)
        num_decoder_layers: int = self.config["num_decoder_layers"]
        num_queries: int = self.config.get("num_queries", 300)
        expansion: float = self.config.get("expansion", 1.0)
        depth_multiplier: float = self.config.get("depth_multiplier", 1.0)
        use_giou: bool = self.config.get("use_giou", True)
        num_denoising: int = self.config.get("num_denoising", 100)
        label_noise_ratio: float = self.config.get("label_noise_ratio", 0.5)
        box_noise_scale: float = self.config.get("box_noise_scale", 1.0)

        self.hidden_dim = hidden_dim
        self.num_queries = num_queries
        self.num_denoising = num_denoising
        self.label_noise_ratio = label_noise_ratio
        self.box_noise_scale = box_noise_scale

        self.backbone.return_channels = self.backbone.return_channels[-3:]
        self.backbone.return_stages = self.backbone.return_stages[-3:]
        self.num_levels = len(self.backbone.return_channels)

        self.encoder = HybridEncoder(
            in_channels=self.backbone.return_channels,
            hidden_dim=hidden_dim,
            num_encoder_layers=num_encoder_layers,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            num_heads=num_heads,
            expansion=expansion,
            depth_multiplier=depth_multiplier,
        )
        self.decoder = RT_DETRDecoder(
            hidden_dim=hidden_dim,
            num_classes=self.num_classes,
            num_queries=self.num_queries,
            num_decoder_layers=num_decoder_layers,
            num_levels=self.num_levels,
            num_heads=num_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            num_decoder_points=num_decoder_points,
        )

        self.matcher = HungarianMatcher(cost_class=2.0, cost_bbox=5.0, cost_giou=2.0, use_giou=use_giou)

        # Denoising class embedding for Contrastive denoising (CDN) training
        if self.num_denoising > 0:
            self.denoising_class_embed = nn.Embedding(self.num_classes + 1, hidden_dim, padding_idx=self.num_classes)

        if self.export_mode is False:
            self.forward = torch.compiler.disable(recursive=False)(self.forward)  # type: ignore[method-assign]

    def _set_cache_enabled(self, enabled: bool) -> None:
        self.encoder.set_cache_enabled(enabled)
        self.decoder.set_cache_enabled(enabled)

    def clear_cache(self) -> None:
        self.encoder.clear_cache()
        self.decoder.clear_cache()

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)
        self.clear_cache()

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        super().set_dynamic_size(dynamic_size)
        self._set_cache_enabled(dynamic_size is False)

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes

        self.decoder.enc_score_head = nn.Linear(self.hidden_dim, num_classes)
        self.decoder.class_embed = nn.ModuleList(
            [nn.Linear(self.hidden_dim, num_classes) for _ in range(len(self.decoder.layers))]
        )

        if self.num_denoising > 0:
            self.denoising_class_embed = nn.Embedding(num_classes + 1, self.hidden_dim, padding_idx=num_classes)

        prior_prob = 0.01
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        nn.init.constant_(self.decoder.enc_score_head.bias, bias_value)
        for class_embed in self.decoder.class_embed:
            nn.init.constant_(class_embed.bias, bias_value)

    def freeze(self, freeze_classifier: bool = True) -> None:
        for param in self.parameters():
            param.requires_grad_(False)

        if freeze_classifier is False:
            for param in self.decoder.class_embed.parameters():
                param.requires_grad_(True)
            for param in self.decoder.enc_score_head.parameters():
                param.requires_grad_(True)
            if self.num_denoising > 0:
                for param in self.denoising_class_embed.parameters():
                    param.requires_grad_(True)

    @staticmethod
    def _get_src_permutation_idx(indices: list[tuple[torch.Tensor, torch.Tensor]]) -> tuple[torch.Tensor, torch.Tensor]:
        batch_idx = torch.concat([torch.full_like(src, i) for i, (src, _) in enumerate(indices)])
        src_idx = torch.concat([src for (src, _) in indices])
        return (batch_idx, src_idx)

    def _class_loss(
        self,
        cls_logits: torch.Tensor,
        box_output: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        indices: list[tuple[torch.Tensor, torch.Tensor]],
        num_boxes: float,
    ) -> torch.Tensor:
        idx = self._get_src_permutation_idx(indices)
        target_classes_o = torch.concat([t["labels"][J] for t, (_, J) in zip(targets, indices)], dim=0)
        target_classes = torch.full(cls_logits.shape[:2], self.num_classes, dtype=torch.int64, device=cls_logits.device)
        target_classes[idx] = target_classes_o

        target_classes_onehot = torch.zeros(
            [cls_logits.shape[0], cls_logits.shape[1], cls_logits.shape[2] + 1],
            dtype=cls_logits.dtype,
            layout=cls_logits.layout,
            device=cls_logits.device,
        )
        target_classes_onehot.scatter_(2, target_classes.unsqueeze(-1), 1)
        target_classes_onehot = target_classes_onehot[:, :, :-1]

        src_boxes = box_output[idx]
        target_boxes = torch.concat([t["boxes"][i] for t, (_, i) in zip(targets, indices)], dim=0)
        ious = torch.diag(
            box_ops.box_iou(
                box_ops.box_convert(src_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
                box_ops.box_convert(target_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
            )
        ).detach()

        target_score_o = torch.zeros(cls_logits.shape[:2], dtype=cls_logits.dtype, device=cls_logits.device)
        target_score_o[idx] = ious.to(cls_logits.dtype)
        target_score = target_score_o.unsqueeze(-1) * target_classes_onehot

        loss = varifocal_loss(cls_logits, target_score, target_classes_onehot, alpha=0.75, gamma=2.0)
        loss_ce = (loss.mean(1).sum() / num_boxes) * cls_logits.shape[1]

        return loss_ce

    def _box_loss(
        self,
        box_output: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        indices: list[tuple[torch.Tensor, torch.Tensor]],
        num_boxes: float,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        idx = self._get_src_permutation_idx(indices)
        src_boxes = box_output[idx]
        target_boxes = torch.concat([t["boxes"][i] for t, (_, i) in zip(targets, indices)], dim=0)

        loss_bbox = F.l1_loss(src_boxes, target_boxes, reduction="none")
        loss_bbox = loss_bbox.sum() / num_boxes

        loss_giou = 1 - torch.diag(
            box_ops.generalized_box_iou(
                box_ops.box_convert(src_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
                box_ops.box_convert(target_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
            )
        )
        loss_giou = loss_giou.sum() / num_boxes

        return (loss_bbox, loss_giou)

    def _compute_denoising_loss(
        self,
        dn_out_bboxes: torch.Tensor,
        dn_out_logits: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        dn_meta: dict[str, Any],
        num_boxes: float,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        dn_positive_idx = dn_meta["dn_positive_idx"]
        num_groups = dn_meta["dn_num_group"]

        loss_ce_list = []
        loss_bbox_list = []
        loss_giou_list = []

        dn_num_boxes = max(num_boxes * num_groups, 1.0)
        for layer_idx in range(dn_out_logits.shape[0]):
            # Construct indices from denoising metadata
            indices = []
            for batch_idx, pos_idx in enumerate(dn_positive_idx):
                if len(pos_idx) > 0:
                    src_idx = pos_idx
                    num_gt = len(targets[batch_idx]["labels"])
                    tgt_idx = torch.arange(num_gt, device=pos_idx.device).repeat(num_groups)
                    indices.append((src_idx, tgt_idx))
                else:
                    indices.append(
                        (
                            torch.tensor([], dtype=torch.long, device=dn_out_logits.device),
                            torch.tensor([], dtype=torch.long, device=dn_out_logits.device),
                        )
                    )

            loss_ce = self._class_loss(
                dn_out_logits[layer_idx], dn_out_bboxes[layer_idx], targets, indices, dn_num_boxes
            )
            loss_bbox, loss_giou = self._box_loss(dn_out_bboxes[layer_idx], targets, indices, dn_num_boxes)

            loss_ce_list.append(loss_ce)
            loss_bbox_list.append(loss_bbox)
            loss_giou_list.append(loss_giou)

        loss_ce_dn = torch.stack(loss_ce_list).sum()
        loss_bbox_dn = torch.stack(loss_bbox_list).sum()
        loss_giou_dn = torch.stack(loss_giou_list).sum()

        return (loss_ce_dn, loss_bbox_dn, loss_giou_dn)

    def _compute_loss_from_outputs(  # pylint: disable=too-many-locals
        self,
        targets: list[dict[str, torch.Tensor]],
        out_bboxes: torch.Tensor,
        out_logits: torch.Tensor,
        enc_topk_bboxes: torch.Tensor,
        enc_topk_logits: torch.Tensor,
        dn_out_bboxes: Optional[torch.Tensor] = None,
        dn_out_logits: Optional[torch.Tensor] = None,
        dn_meta: Optional[dict[str, Any]] = None,
    ) -> dict[str, torch.Tensor]:
        # Compute the average number of target boxes across all nodes
        num_boxes = sum(len(t["labels"]) for t in targets)
        num_boxes = torch.as_tensor([num_boxes], dtype=torch.float, device=out_logits.device)
        if training_utils.is_dist_available_and_initialized() is True:
            torch.distributed.all_reduce(num_boxes)

        num_boxes = torch.clamp(num_boxes / training_utils.get_world_size(), min=1)

        loss_ce_list = []
        loss_bbox_list = []
        loss_giou_list = []

        # Decoder losses (all layers)
        for layer_idx in range(out_logits.shape[0]):
            indices = self.matcher(out_logits[layer_idx], out_bboxes[layer_idx], targets)
            loss_ce = self._class_loss(out_logits[layer_idx], out_bboxes[layer_idx], targets, indices, num_boxes)
            loss_bbox, loss_giou = self._box_loss(out_bboxes[layer_idx], targets, indices, num_boxes)
            loss_ce_list.append(loss_ce)
            loss_bbox_list.append(loss_bbox)
            loss_giou_list.append(loss_giou)

        # Encoder auxiliary loss
        enc_indices = self.matcher(enc_topk_logits, enc_topk_bboxes, targets)
        loss_ce_enc = self._class_loss(enc_topk_logits, enc_topk_bboxes, targets, enc_indices, num_boxes)
        loss_bbox_enc, loss_giou_enc = self._box_loss(enc_topk_bboxes, targets, enc_indices, num_boxes)
        loss_ce_list.append(loss_ce_enc)
        loss_bbox_list.append(loss_bbox_enc)
        loss_giou_list.append(loss_giou_enc)

        loss_ce = torch.stack(loss_ce_list).sum()  # VFL weight is 1
        loss_bbox = torch.stack(loss_bbox_list).sum() * 5
        loss_giou = torch.stack(loss_giou_list).sum() * 2

        # Add denoising loss if available
        if dn_out_bboxes is not None and dn_out_logits is not None and dn_meta is not None:
            loss_ce_dn, loss_bbox_dn, loss_giou_dn = self._compute_denoising_loss(
                dn_out_bboxes, dn_out_logits, targets, dn_meta, num_boxes
            )
            loss_ce = loss_ce + loss_ce_dn
            loss_bbox = loss_bbox + loss_bbox_dn * 5
            loss_giou = loss_giou + loss_giou_dn * 2

        losses = {
            "labels": loss_ce,
            "boxes": loss_bbox,
            "giou": loss_giou,
        }

        return losses

    @torch.jit.unused  # type: ignore[untyped-decorator]
    @torch.compiler.disable()  # type: ignore[untyped-decorator]
    def compute_loss(
        self,
        encoder_features: list[torch.Tensor],
        spatial_shapes: list[list[int]],
        level_start_index: list[int],
        targets: list[dict[str, torch.Tensor]],
        images: Any,
        masks: Optional[list[torch.Tensor]] = None,
    ) -> dict[str, torch.Tensor]:
        for idx, target in enumerate(targets):
            boxes = target["boxes"]
            boxes = box_ops.box_convert(boxes, in_fmt="xyxy", out_fmt="cxcywh")
            scale = images.image_sizes[idx].flip(0).repeat(2).float()  # flip to [W, H], repeat to [W, H, W, H]
            boxes = boxes / scale
            targets[idx]["boxes"] = boxes
            targets[idx]["labels"] = target["labels"] - 1  # No background

        denoising_class, denoising_bbox_unact, attn_mask, dn_meta = self._prepare_cdn_queries(targets)

        out_bboxes, out_logits, enc_topk_bboxes, enc_topk_logits = self.decoder(
            encoder_features,
            spatial_shapes,
            level_start_index,
            denoising_class,
            denoising_bbox_unact,
            attn_mask,
            masks,
        )

        if dn_meta is not None:
            dn_num_split, _num_queries = dn_meta["dn_num_split"]
            dn_out_bboxes = out_bboxes[:, :, :dn_num_split]
            dn_out_logits = out_logits[:, :, :dn_num_split]
            out_bboxes = out_bboxes[:, :, dn_num_split:]
            out_logits = out_logits[:, :, dn_num_split:]
        else:
            dn_out_bboxes = None
            dn_out_logits = None

        losses: dict[str, torch.Tensor] = self._compute_loss_from_outputs(
            targets, out_bboxes, out_logits, enc_topk_bboxes, enc_topk_logits, dn_out_bboxes, dn_out_logits, dn_meta
        )

        return losses

    def postprocess_detections(
        self, class_logits: torch.Tensor, box_regression: torch.Tensor, image_sizes: torch.Tensor
    ) -> list[dict[str, torch.Tensor]]:
        prob = class_logits.sigmoid()
        topk_values, topk_indexes = torch.topk(prob.view(class_logits.shape[0], -1), k=self.decoder.num_queries, dim=1)
        scores = topk_values
        topk_boxes = topk_indexes // class_logits.shape[2]
        labels = topk_indexes % class_logits.shape[2]
        labels += 1  # Background offset

        # Convert to [x0, y0, x1, y1] format
        boxes = box_ops.box_convert(box_regression, in_fmt="cxcywh", out_fmt="xyxy")
        boxes = torch.gather(boxes, 1, topk_boxes.unsqueeze(-1).expand(-1, -1, 4))

        # Convert from relative [0, 1] to absolute [0, height] coordinates
        img_h, img_w = image_sizes.unbind(1)
        scale_fct = torch.stack([img_w, img_h, img_w, img_h], dim=1)
        boxes = boxes * scale_fct[:, None, :]

        detections: list[dict[str, torch.Tensor]] = []
        for s, l, b in zip(scores, labels, boxes):
            detections.append(
                {
                    "boxes": b,
                    "scores": s,
                    "labels": l,
                }
            )

        return detections

    @torch.jit.unused  # type: ignore[untyped-decorator]
    def _prepare_cdn_queries(
        self, targets: list[dict[str, torch.Tensor]]
    ) -> tuple[Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor], Optional[dict[str, Any]]]:
        if self.num_denoising > 0:
            result: tuple[
                Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor], Optional[dict[str, Any]]
            ] = get_contrastive_denoising_training_group(
                targets,
                self.num_classes,
                self.num_queries,
                self.denoising_class_embed,
                num_denoising_queries=self.num_denoising,
                label_noise_ratio=self.label_noise_ratio,
                box_noise_scale=self.box_noise_scale,
            )
            return result

        return (None, None, None, None)

    def forward_net(
        self, x: torch.Tensor, masks: Optional[torch.Tensor]
    ) -> tuple[list[torch.Tensor], Optional[list[torch.Tensor]]]:
        features: dict[str, torch.Tensor] = self.backbone.detection_features(x)
        feature_list = list(features.values())

        mask_list: Optional[list[torch.Tensor]] = None
        if masks is not None:
            mask_list = []
            for feat in feature_list:
                m = F.interpolate(masks[None].float(), size=feat.shape[-2:], mode="nearest").to(torch.bool)[0]
                mask_list.append(m)

        encoder_features = self.encoder(feature_list, masks=mask_list)

        return (encoder_features, mask_list)

    def forward(
        self,
        x: torch.Tensor,
        targets: Optional[list[dict[str, torch.Tensor]]] = None,
        masks: Optional[torch.Tensor] = None,
        image_sizes: Optional[list[tuple[int, int]]] = None,
    ) -> tuple[list[dict[str, torch.Tensor]], dict[str, torch.Tensor]]:
        self._input_check(targets)
        images = self._to_img_list(x, image_sizes)

        encoder_features, mask_list = self.forward_net(x, masks)

        # Prepare spatial shapes and level start index
        spatial_shapes: list[list[int]] = []
        level_start_index: list[int] = [0]
        for feat in encoder_features:
            H = feat.shape[2]
            W = feat.shape[3]
            spatial_shapes.append([H, W])
            level_start_index.append(H * W + level_start_index[-1])

        level_start_index.pop()

        detections: list[dict[str, torch.Tensor]] = []
        losses: dict[str, torch.Tensor] = {}
        if self.training is True:
            assert targets is not None, "targets should not be None when in training mode"
            losses = self.compute_loss(encoder_features, spatial_shapes, level_start_index, targets, images, mask_list)
        else:
            # Inference path - no CDN
            out_bboxes, out_logits, _, _ = self.decoder(
                encoder_features, spatial_shapes, level_start_index, padding_mask=mask_list, return_intermediates=False
            )
            detections = self.postprocess_detections(out_logits, out_bboxes, images.image_sizes)

        return (detections, losses)

    @torch.no_grad()  # type: ignore[untyped-decorator]
    def reparameterize_model(self) -> None:
        if self.reparameterized is True:
            return

        for module in self.modules():
            if hasattr(module, "reparameterize") is True:
                module.reparameterize()

        self.reparameterized = True


registry.register_model_config(
    "rt_detr_v1_t", RT_DETR_v1, config={"num_decoder_layers": 3, "expansion": 0.5, "depth_multiplier": 0.33}
)
registry.register_model_config("rt_detr_v1_s", RT_DETR_v1, config={"num_decoder_layers": 3, "expansion": 0.5})
registry.register_model_config("rt_detr_v1", RT_DETR_v1, config={"num_decoder_layers": 6})



"""
Gradient-weighted Class Activation Mapping (Grad-CAM), adapted from
https://github.com/jacobgil/pytorch-grad-cam

Paper "Grad-CAM: Visual Explanations from Deep Networks via Gradient-based Localization",
https://arxiv.org/abs/1610.02391
"""

# Reference license: MIT

from collections.abc import Callable
from pathlib import Path
from typing import Optional

import numpy as np
import numpy.typing as npt
import torch
from PIL import Image
from torch import nn
from torch.utils.hooks import RemovableHandle

from birder.introspection.base import InterpretabilityResult
from birder.introspection.base import predict_class
from birder.introspection.base import preprocess_image
from birder.introspection.base import scale_cam_image
from birder.introspection.base import show_mask_on_image
from birder.introspection.base import validate_target_class


def compute_cam(activations: npt.NDArray[np.float32], gradients: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
    weights: npt.NDArray[np.float32] = np.mean(gradients, axis=(2, 3))
    weighted_activations = weights[:, :, None, None] * activations
    cam: npt.NDArray[np.float32] = weighted_activations.sum(axis=1)
    cam = np.maximum(cam, 0)

    return cam


class ActivationCapture:
    def __init__(
        self,
        model: nn.Module,
        target_layer: nn.Module,
        reshape_transform: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
    ) -> None:
        self.model = model
        self.target_layer = target_layer
        self.reshape_transform = reshape_transform

        self.activations: Optional[torch.Tensor] = None
        self.gradients: Optional[torch.Tensor] = None
        self.handles: list[RemovableHandle] = []

        # Register hooks
        self.handles.append(target_layer.register_forward_hook(self._save_activation))
        self.handles.append(target_layer.register_forward_hook(self._save_gradient))

    def _save_activation(self, _module: nn.Module, _input: torch.Tensor, output: torch.Tensor) -> None:
        if self.reshape_transform is not None:
            output = self.reshape_transform(output)

        self.activations = output.cpu().detach()

    def _save_gradient(self, _module: nn.Module, _input: torch.Tensor, output: torch.Tensor) -> None:
        if hasattr(output, "requires_grad") is False or output.requires_grad is False:
            return

        def _store_grad(grad: torch.Tensor) -> None:
            if self.reshape_transform is not None:
                grad = self.reshape_transform(grad)

            self.gradients = grad.cpu().detach()

        output.register_hook(_store_grad)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)

    def release(self) -> None:
        for handle in self.handles:
            handle.remove()


class GradCAM:
    def __init__(
        self,
        net: nn.Module,
        device: torch.device,
        transform: Callable[..., torch.Tensor],
        target_layer: nn.Module,
        reshape_transform: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
    ) -> None:
        self.net = net.eval()
        self.device = device
        self.transform = transform
        self.target_layer = target_layer

        self.activation_capture = ActivationCapture(net, target_layer, reshape_transform)

    def __call__(self, image: str | Path | Image.Image, target_class: Optional[int] = None) -> InterpretabilityResult:
        input_tensor, rgb_img = preprocess_image(image, self.transform, self.device)
        input_tensor.requires_grad_(True)

        # Forward pass
        logits = self.activation_capture(input_tensor)

        # Determine target class
        if target_class is None:
            target_class = predict_class(logits)
        else:
            validate_target_class(target_class, logits.shape[-1])

        # Backward pass
        self.net.zero_grad()
        loss = logits[0, target_class]
        loss.backward(retain_graph=False)

        # Get captured activations and gradients
        if self.activation_capture.activations is None:
            raise RuntimeError("No activations captured")

        if self.activation_capture.gradients is None:
            raise RuntimeError("No gradients captured")

        activations = self.activation_capture.activations.numpy()
        gradients = self.activation_capture.gradients.numpy()

        # Compute CAM
        cam = compute_cam(activations, gradients)
        target_size = (input_tensor.size(-1), input_tensor.size(-2))
        cam_scaled = scale_cam_image(cam, target_size)
        grayscale_cam = cam_scaled[0]

        # Create visualization
        visualization = show_mask_on_image(rgb_img, grayscale_cam)

        return InterpretabilityResult(
            original_image=rgb_img,
            visualization=visualization,
            raw_output=grayscale_cam,
            logits=logits.detach(),
            predicted_class=target_class,
        )

    def __del__(self) -> None:
        if hasattr(self, "activation_capture") is True:
            self.activation_capture.release()



"""
AwA2 benchmark using MLP probe for multi-label attribute prediction

Paper "Zero-Shot Learning -- A Comprehensive Evaluation of the Good, the Bad and the Ugly"
https://arxiv.org/abs/1707.00600
"""

import argparse
import logging
import time
from pathlib import Path
from typing import Any

import numpy as np
import numpy.typing as npt
import polars as pl
import torch
from rich.console import Console
from rich.table import Table
from sklearn.metrics import f1_score
from torchvision.datasets import ImageFolder

from birder.common import cli
from birder.common import lib
from birder.conf import settings
from birder.datahub.evaluation import AwA2
from birder.eval._embeddings import load_embeddings
from birder.eval.methods.mlp import evaluate_mlp
from birder.eval.methods.mlp import train_mlp

logger = logging.getLogger(__name__)


def _print_summary_table(results: list[dict[str, Any]]) -> None:
    console = Console()

    table = Table(show_header=True, header_style="bold dark_magenta")
    table.add_column("AwA2 (MLP)", style="dim")
    table.add_column("Macro F1", justify="right")
    table.add_column("Std", justify="right")
    table.add_column("Runs", justify="right")

    for result in results:
        table.add_row(
            Path(result["embeddings_file"]).name,
            f"{result['macro_f1']:.4f}",
            f"{result['macro_f1_std']:.4f}",
            f"{result['num_runs']}",
        )

    console.print(table)


def _write_results_csv(results: list[dict[str, Any]], attribute_names: list[str], output_path: Path) -> None:
    rows: list[dict[str, Any]] = []
    for result in results:
        row: dict[str, Any] = {
            "embeddings_file": result["embeddings_file"],
            "method": result["method"],
            "metric_mode": result["metric_mode"],
            "macro_f1": result["macro_f1"],
            "macro_f1_std": result["macro_f1_std"],
            "num_runs": result["num_runs"],
        }
        for attr in attribute_names:
            row[f"f1_{attr}"] = result["per_attribute_f1"].get(attr)

        rows.append(row)

    pl.DataFrame(rows).write_csv(output_path)
    logger.info(f"Results saved to {output_path}")


def _load_awa2_metadata(dataset: AwA2) -> tuple[pl.DataFrame, npt.NDArray[np.float32], list[str], list[str], list[str]]:
    """
    Load AwA2 metadata: image paths, class assignments, and attribute matrix.

    Returns
    -------
    metadata_df
        DataFrame with columns: id (image stem), class_name
    attribute_matrix
        Binary attribute matrix of shape (num_classes, num_attributes)
    class_names
        List of class names in order
    train_classes
        List of training class names
    test_classes
        List of test class names
    """

    # Load class names (1-indexed in file)
    class_names: list[str] = []
    with open(dataset.classes_path, encoding="utf-8") as f:
        for line in f:
            parts = line.strip().split("\t")
            class_names.append(parts[1])

    # Load attribute matrix (one row per class, 85 attributes)
    attribute_matrix = np.loadtxt(dataset.predicate_matrix_binary_path, dtype=np.float32)

    # Load train/test class split
    with open(dataset.trainclasses_path, encoding="utf-8") as f:
        train_classes = [line.strip() for line in f if line.strip()]

    with open(dataset.testclasses_path, encoding="utf-8") as f:
        test_classes = [line.strip() for line in f if line.strip()]

    # Load image paths using ImageFolder
    image_dataset = ImageFolder(str(dataset.images_dir))
    rows: list[dict[str, Any]] = []
    for path, class_idx in image_dataset.samples:
        class_name = image_dataset.classes[class_idx]
        rows.append({"id": Path(path).stem, "class_name": class_name})

    metadata_df = pl.DataFrame(rows)

    return (metadata_df, attribute_matrix, class_names, train_classes, test_classes)


def _load_embeddings_with_labels(
    embeddings_path: str,
    metadata_df: pl.DataFrame,
    attribute_matrix: npt.NDArray[np.float32],
    class_names: list[str],
    train_classes: list[str],
    test_classes: list[str],
) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.float32], npt.NDArray[np.float32], npt.NDArray[np.float32]]:
    logger.info(f"Loading embeddings from {embeddings_path}")
    emb_df = load_embeddings(embeddings_path)

    # Join embeddings with metadata
    joined = metadata_df.join(emb_df, on="id", how="inner")
    if joined.height < metadata_df.height:
        logger.warning(f"Join dropped {metadata_df.height - joined.height} samples (missing embeddings)")

    # Create class name to index mapping
    class_to_idx = {name: idx for idx, name in enumerate(class_names)}

    # Split into train/test based on class membership
    train_mask = joined.get_column("class_name").is_in(train_classes)
    test_mask = joined.get_column("class_name").is_in(test_classes)

    train_data = joined.filter(train_mask)
    test_data = joined.filter(test_mask)

    # Extract features
    x_train = train_data.get_column("embedding").to_numpy().astype(np.float32, copy=False)
    x_test = test_data.get_column("embedding").to_numpy().astype(np.float32, copy=False)

    # Get labels from attribute matrix (class-level attributes)
    train_class_indices = [class_to_idx[name] for name in train_data.get_column("class_name").to_list()]
    test_class_indices = [class_to_idx[name] for name in test_data.get_column("class_name").to_list()]

    y_train = attribute_matrix[train_class_indices]
    y_test = attribute_matrix[test_class_indices]

    logger.info(f"Train: {x_train.shape[0]} samples ({len(train_classes)} classes)")
    logger.info(f"Test: {x_test.shape[0]} samples ({len(test_classes)} classes)")
    logger.info(f"Features: {x_train.shape[1]} dims, Attributes: {attribute_matrix.shape[1]}")

    return (x_train, y_train, x_test, y_test)


def _compute_macro_f1(
    y_true: npt.NDArray[np.int_], y_pred: npt.NDArray[np.int_], metric_mode: str
) -> tuple[float, int]:
    if metric_mode == "all":
        num_attrs = y_true.shape[1]
        score = f1_score(y_true, y_pred, average="macro", zero_division=0.0)
        return (float(score), num_attrs)

    if metric_mode == "present-only":
        present_attrs = np.where(y_true.sum(axis=0) > 0)[0]
        if len(present_attrs) == 0:
            logger.warning("No positive attributes in y_true, falling back to --metric-mode all")
            score = f1_score(y_true, y_pred, average="macro", zero_division=0.0)
            return (float(score), y_true.shape[1])

        score = f1_score(y_true, y_pred, average="macro", labels=present_attrs, zero_division=0.0)
        return (float(score), int(len(present_attrs)))

    raise ValueError(f"Unsupported metric mode: {metric_mode}")


# pylint: disable=too-many-locals
def evaluate_awa2_single(
    x_train: npt.NDArray[np.float32],
    y_train: npt.NDArray[np.float32],
    x_test: npt.NDArray[np.float32],
    y_test: npt.NDArray[np.float32],
    attribute_names: list[str],
    args: argparse.Namespace,
    embeddings_path: str,
    device: torch.device,
) -> dict[str, Any]:
    num_attributes = len(attribute_names)

    scores: list[float] = []
    per_attribute_f1_runs: list[dict[str, float]] = []

    for run in range(args.runs):
        run_seed = args.seed + run
        logger.info(f"Run {run + 1}/{args.runs} (seed={run_seed})")

        # Train MLP
        model = train_mlp(
            x_train,
            y_train,
            num_classes=num_attributes,
            device=device,
            epochs=args.epochs,
            batch_size=args.batch_size,
            lr=args.lr,
            hidden_dim=args.hidden_dim,
            dropout=args.dropout,
            seed=run_seed,
        )

        # Evaluate
        y_pred, y_true, _ = evaluate_mlp(model, x_test, y_test, batch_size=args.batch_size, device=device)
        macro_f1, num_attrs_scored = _compute_macro_f1(y_true, y_pred, args.metric_mode)
        scores.append(macro_f1)

        # Per-attribute F1
        per_attribute_f1: dict[str, float] = {}
        for i, attr in enumerate(attribute_names):
            attr_f1 = f1_score(y_true[:, i], y_pred[:, i], average="binary", zero_division=0.0)
            per_attribute_f1[attr] = float(attr_f1)

        per_attribute_f1_runs.append(per_attribute_f1)
        logger.info(
            f"Run {run + 1}/{args.runs} - Macro F1 ({args.metric_mode}, {num_attrs_scored} attrs): {macro_f1:.4f}"
        )

    # Average results
    scores_arr = np.array(scores)
    mean_f1 = float(scores_arr.mean())
    std_f1 = float(scores_arr.std(ddof=1)) if len(scores) > 1 else 0.0

    # Average per-attribute F1 across runs
    avg_per_attribute_f1: dict[str, float] = {}
    for attr in attribute_names:
        attr_scores = [run_f1[attr] for run_f1 in per_attribute_f1_runs]
        avg_per_attribute_f1[attr] = float(np.mean(attr_scores))

    logger.info(f"Mean Macro F1 over {args.runs} runs: {mean_f1:.4f} +/- {std_f1:.4f} (std)")

    return {
        "method": "mlp",
        "metric_mode": args.metric_mode,
        "macro_f1": mean_f1,
        "macro_f1_std": std_f1,
        "num_runs": args.runs,
        "per_attribute_f1": avg_per_attribute_f1,
        "embeddings_file": str(embeddings_path),
    }


def evaluate_awa2(args: argparse.Namespace) -> None:
    tic = time.time()

    if args.gpu is True:
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    if args.gpu_id is not None:
        torch.cuda.set_device(args.gpu_id)

    logger.info(f"Using device {device}")
    logger.info(f"Loading AwA2 dataset from {args.dataset_path}")
    logger.info(f"Metric mode: {args.metric_mode}")
    dataset = AwA2(args.dataset_path)
    attribute_names = dataset.attribute_names

    metadata_df, attribute_matrix, class_names, train_classes, test_classes = _load_awa2_metadata(dataset)
    logger.info(f"Loaded metadata for {metadata_df.height} images")
    logger.info(f"Train classes: {len(train_classes)}, Test classes: {len(test_classes)}")

    results: list[dict[str, Any]] = []
    total = len(args.embeddings)
    for idx, embeddings_path in enumerate(args.embeddings, start=1):
        logger.info(f"Processing embeddings {idx}/{total}: {embeddings_path}")
        x_train, y_train, x_test, y_test = _load_embeddings_with_labels(
            embeddings_path, metadata_df, attribute_matrix, class_names, train_classes, test_classes
        )

        result = evaluate_awa2_single(x_train, y_train, x_test, y_test, attribute_names, args, embeddings_path, device)
        results.append(result)

    _print_summary_table(results)

    if args.dry_run is False:
        output_dir = settings.RESULTS_DIR.joinpath(args.dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        output_path = output_dir.joinpath("awa2.csv")
        _write_results_csv(results, attribute_names, output_path)

    toc = time.time()
    logger.info(f"AwA2 benchmark completed in {lib.format_duration(toc - tic)}")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "awa2",
        allow_abbrev=False,
        help="run AwA2 benchmark - 85 attribute multi-label classification using MLP probe",
        description="run AwA2 benchmark - 85 attribute multi-label classification using MLP probe",
        epilog=(
            "Usage examples:\n"
            "python -m birder.eval awa2 --embeddings "
            "results/awa2_embeddings.parquet "
            "--dataset-path ~/Datasets/Animals_with_Attributes2 --dry-run\n"
            "python -m birder.eval awa2 --embeddings results/awa2_*.parquet "
            "--dataset-path ~/Datasets/Animals_with_Attributes2 --gpu\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "--embeddings", type=str, nargs="+", metavar="FILE", help="paths to embeddings parquet files"
    )
    subparser.add_argument("--dataset-path", type=str, metavar="PATH", help="path to AwA2 dataset root")
    subparser.add_argument("--runs", type=int, default=3, help="number of evaluation runs")
    subparser.add_argument("--epochs", type=int, default=100, help="training epochs per run")
    subparser.add_argument("--batch-size", type=int, default=128, help="batch size for training and inference")
    subparser.add_argument("--lr", type=float, default=1e-4, help="learning rate")
    subparser.add_argument("--hidden-dim", type=int, default=512, help="MLP hidden layer dimension")
    subparser.add_argument("--dropout", type=float, default=0.5, help="dropout probability")
    subparser.add_argument("--seed", type=int, default=0, help="base random seed")
    subparser.add_argument(
        "--metric-mode",
        type=str,
        choices=["all", "present-only"],
        default="present-only",
        help="macro F1 mode: all attributes or only attributes present in test split",
    )
    subparser.add_argument("--gpu", default=False, action="store_true", help="use gpu")
    subparser.add_argument("--gpu-id", type=int, metavar="ID", help="gpu id to use")
    subparser.add_argument(
        "--dir", type=str, default="awa2", help="place all outputs in a sub-directory (relative to results)"
    )
    subparser.add_argument("--dry-run", default=False, action="store_true", help="skip saving results to file")
    subparser.set_defaults(func=main)


def validate_args(args: argparse.Namespace) -> None:
    if args.embeddings is None:
        raise cli.ValidationError("--embeddings is required")
    if args.dataset_path is None:
        raise cli.ValidationError("--dataset-path is required")


def main(args: argparse.Namespace) -> None:
    validate_args(args)
    evaluate_awa2(args)



from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
import torch.amp
from PIL import Image
from torch.nn import functional as F
from torch.utils.data import DataLoader
from tqdm import tqdm

from birder.conf import settings
from birder.data.collators.detection import batch_images
from birder.data.transforms.detection import InferenceTransform
from birder.inference.wbf import fuse_detections_wbf
from birder.net.base import make_divisible


def _normalize_image_sizes(inputs: torch.Tensor, image_sizes: Optional[list[tuple[int, int]]]) -> list[tuple[int, int]]:
    if image_sizes is not None:
        return image_sizes

    _, _, height, width = inputs.shape
    return [(height, width) for _ in range(inputs.size(0))]


def _hflip_inputs(inputs: torch.Tensor, image_sizes: list[tuple[int, int]]) -> torch.Tensor:
    # Detection collator pads on the right/bottom, so flip only the valid region to keep padding aligned.
    flipped = inputs.clone()
    for idx, (height, width) in enumerate(image_sizes):
        flipped[idx, :, :height, :width] = torch.flip(inputs[idx, :, :height, :width], dims=[2])

    return flipped


def _resize_batch(
    inputs: torch.Tensor, image_sizes: list[tuple[int, int]], scale: float, size_divisible: int
) -> tuple[torch.Tensor, torch.Tensor, list[tuple[int, int]]]:
    resized_images: list[torch.Tensor] = []
    for idx, (height, width) in enumerate(image_sizes):
        target_h = make_divisible(height * scale, size_divisible)
        target_w = make_divisible(width * scale, size_divisible)
        image = inputs[idx, :, :height, :width]
        resized = F.interpolate(image.unsqueeze(0), size=(target_h, target_w), mode="bilinear", align_corners=False)
        resized_images.append(resized.squeeze(0))

    return batch_images(resized_images, size_divisible)


def _rescale_boxes(boxes: torch.Tensor, from_size: tuple[int, int], to_size: tuple[int, int]) -> torch.Tensor:
    scale_w = to_size[1] / from_size[1]
    scale_h = to_size[0] / from_size[0]
    scale = boxes.new_tensor([scale_w, scale_h, scale_w, scale_h])
    return boxes * scale


def _rescale_detections(
    detections: list[dict[str, torch.Tensor]],
    from_sizes: list[tuple[int, int]],
    to_sizes: list[tuple[int, int]],
) -> list[dict[str, torch.Tensor]]:
    for idx, (detection, from_size, to_size) in enumerate(zip(detections, from_sizes, to_sizes)):
        boxes = detection["boxes"]
        if boxes.numel() == 0:
            continue

        detections[idx]["boxes"] = _rescale_boxes(boxes, from_size, to_size)

    return detections


def _invert_hflip_boxes(boxes: torch.Tensor, image_size: tuple[int, int]) -> torch.Tensor:
    width = boxes.new_tensor(image_size[1])
    x1 = boxes[:, 0]
    x2 = boxes[:, 2]
    flipped = boxes.clone()
    flipped[:, 0] = width - x2
    flipped[:, 2] = width - x1

    return flipped


def _invert_detections(
    detections: list[dict[str, torch.Tensor]], image_sizes: list[tuple[int, int]]
) -> list[dict[str, torch.Tensor]]:
    for idx, (detection, image_size) in enumerate(zip(detections, image_sizes)):
        boxes = detection["boxes"]
        if boxes.numel() == 0:
            continue

        detections[idx]["boxes"] = _invert_hflip_boxes(boxes, image_size)

    return detections


def infer_image(
    net: torch.nn.Module | torch.ScriptModule,
    sample: Image.Image | str,
    transform: Callable[..., torch.Tensor],
    tta: bool = False,
    device: Optional[torch.device] = None,
    score_threshold: Optional[float] = None,
    **kwargs: Any,
) -> dict[str, torch.Tensor]:
    """
    Perform inference on a single image

    This convenience function allows for quick, one-off detection of an image.

    Raises
    ------
    TypeError
        If the sample is neither a string nor a PIL Image object.
    """

    image: Image.Image
    if isinstance(sample, str):
        image = Image.open(sample)
    elif isinstance(sample, Image.Image):
        image = sample
    else:
        raise TypeError("Unknown sample type")

    if device is None:
        device = torch.device("cpu")

    input_tensor = transform(image).unsqueeze(dim=0).to(device)
    detections = infer_batch(net, input_tensor, tta=tta, **kwargs)
    if score_threshold is not None:
        for i, detection in enumerate(detections):
            idxs = torch.where(detection["scores"] > score_threshold)
            detections[i]["scores"] = detection["scores"][idxs]
            detections[i]["boxes"] = detection["boxes"][idxs]
            detections[i]["labels"] = detection["labels"][idxs]

    detections = InferenceTransform.postprocess(
        detections, [input_tensor.shape[2:]], [image.size[::-1]]  # type: ignore[list-item]
    )

    return detections[0]


def infer_batch(
    net: torch.nn.Module | torch.ScriptModule,
    inputs: torch.Tensor,
    masks: Optional[torch.Tensor] = None,
    image_sizes: Optional[list[tuple[int, int]]] = None,
    tta: bool = False,
    channels_last: bool = False,
    **kwargs: Any,
) -> list[dict[str, torch.Tensor]]:
    # The 'inputs' are assumed to be in the correct memory format

    if tta is False:
        detections, _ = net(inputs, masks=masks, image_sizes=image_sizes, **kwargs)
        return detections  # type: ignore[no-any-return]

    normalized_sizes = _normalize_image_sizes(inputs, image_sizes)
    detections_list: list[list[dict[str, torch.Tensor]]] = []

    for scale in (0.8, 1.0, 1.2):
        scaled_inputs, scaled_masks, scaled_sizes = _resize_batch(inputs, normalized_sizes, scale, size_divisible=32)
        if channels_last is True:
            scaled_inputs = scaled_inputs.to(memory_format=torch.channels_last)

        detections, _ = net(scaled_inputs, masks=scaled_masks, image_sizes=scaled_sizes, **kwargs)
        detections = _rescale_detections(detections, scaled_sizes, normalized_sizes)
        detections_list.append(detections)

        flipped_inputs = _hflip_inputs(scaled_inputs, scaled_sizes)
        if channels_last is True:
            flipped_inputs = flipped_inputs.to(memory_format=torch.channels_last)

        flipped_detections, _ = net(flipped_inputs, masks=scaled_masks, image_sizes=scaled_sizes, **kwargs)
        flipped_detections = _invert_detections(flipped_detections, scaled_sizes)
        flipped_detections = _rescale_detections(flipped_detections, scaled_sizes, normalized_sizes)
        detections_list.append(flipped_detections)

    return fuse_detections_wbf(detections_list, iou_thr=0.55, conf_type="avg")


def infer_dataloader(
    device: torch.device,
    net: torch.nn.Module | torch.ScriptModule,
    dataloader: DataLoader,
    tta: bool = False,
    channels_last: bool = False,
    model_dtype: torch.dtype = torch.float32,
    amp: bool = False,
    amp_dtype: Optional[torch.dtype] = None,
    num_samples: Optional[int] = None,
    batch_callback: Optional[
        Callable[
            [list[str], torch.Tensor, list[dict[str, torch.Tensor]], list[dict[str, Any]], list[tuple[int, int]]], None
        ]
    ] = None,
) -> tuple[list[str], list[dict[str, torch.Tensor]], list[dict[str, Any]]]:
    """
    Perform inference on a DataLoader using a given neural network.

    This function runs inference on a dataset provided through a DataLoader,
    optionally using mixed precision (amp).
    All returned detections and targets are transformed to original
    image coordinates regardless of the model's inference resolution.

    Parameters
    ----------
    device
        The device to run the inference on.
    net
        The model to use for inference.
    dataloader
        The DataLoader containing the dataset to perform inference on.
    tta
        Run inference with multi-scale and horizontal flip test time augmentation and fuse results with WBF.
    channels_last
        If True, convert input batches to channels-last memory format before inference.
    model_dtype
        The base dtype to use.
    amp
        Whether to use automatic mixed precision.
    amp_dtype
        The mixed precision dtype.
    num_samples
        The total number of samples in the dataloader.
    batch_callback
        A function to be called after each batch is processed. If provided, it
        should accept four arguments:
        - list[str]: A list of file paths for the current batch
        - torch.Tensor: The input tensor for the current batch
        - list[dict[str, torch.Tensor]]: The detections for the current batch
        - list[dict[str, Any]]: A list of targets for the current batch
        - list[tuple[int, int]]: The image sizes for the current batch

    Returns
    -------
    A tuple containing three elements:
    - list[str]: A list of all processed file paths.
    - list[dict[str, torch.Tensor]]: A list of detection dictionaries.
    - list[dict[str, Any]]: A list of all targets.

    Notes
    -----
    - The function uses a progress bar (tqdm) to show the inference progress.
    - If 'num_samples' is not provided, the progress bar may not accurately
      reflect the total number of samples processed.
    - The batch_callback, if provided, is called after each batch is processed,
      allowing for real-time analysis or logging of results.
    """

    net.to(device, dtype=model_dtype)
    detections_list: list[dict[str, torch.Tensor]] = []
    target_list: list[dict[str, Any]] = []
    sample_paths: list[str] = []
    batch_size = dataloader.batch_size
    with tqdm(total=num_samples, initial=0, unit="images", unit_scale=True, leave=False) as progress:
        for file_paths, inputs, targets, orig_sizes, masks, image_sizes in dataloader:
            # Inference
            if channels_last is True:
                inputs = inputs.to(device, dtype=model_dtype, non_blocking=True, memory_format=torch.channels_last)
            else:
                inputs = inputs.to(device, dtype=model_dtype, non_blocking=True)

            masks = masks.to(device, non_blocking=True)

            with torch.amp.autocast(device.type, enabled=amp, dtype=amp_dtype):
                detections = infer_batch(
                    net, inputs, masks=masks, image_sizes=image_sizes, tta=tta, channels_last=channels_last
                )

            detections = InferenceTransform.postprocess(detections, image_sizes, orig_sizes)
            if targets[0] != settings.NO_LABEL:
                targets = InferenceTransform.postprocess(targets, image_sizes, orig_sizes)

            detections_list.extend(detections)

            # Set targets and sample list
            target_list.extend(targets)
            sample_paths.extend(file_paths)

            if batch_callback is not None:
                batch_callback(file_paths, inputs, detections, targets, image_sizes)

            # Update progress bar
            progress.update(n=batch_size)

    return (sample_paths, detections_list, target_list)



"""
RoPE ViT model configuration registrations

This file contains *only* model variant definitions and their registration
with the global model registry. The actual RoPE ViT implementation lives in rope_vit.py.
"""

from birder.model_registry import registry
from birder.net._vit_configs import BASE
from birder.net._vit_configs import GIANT
from birder.net._vit_configs import GIGANTIC
from birder.net._vit_configs import HUGE
from birder.net._vit_configs import LARGE
from birder.net._vit_configs import MEDIUM
from birder.net._vit_configs import SMALL
from birder.net._vit_configs import SO150
from birder.net._vit_configs import SO400
from birder.net._vit_configs import TINY
from birder.net.base import BaseNet

# Vision Transformer Model Naming Convention
# ==========================================
#
# Model names follow a structured pattern to encode architectural choices:
# [rope_]vit_[reg{N}_][size][patch_size][_components][_pooling][_c{N}]
#
# Core Components:
# - rope_       : Rotary Position Embedding (RoPE) enabled
# - rope_i_     : Rotary Position Embedding (RoPE) enabled with interleaved rotation - implies different temp, indexing
# - vit_        : Vision Transformer base architecture
# - reg{N}_     : Register tokens (N = number of register tokens, e.g., reg4, reg8)
# - size        : Model size (s=small, b=base, l=large, or specific like so150m)
# - patch_size  : Patch size (e.g., 14, 16, 32 for 14x14, 16x16, 32x32 patches)
#
# Optional Components:
#     Position Embeddings:
#     - nps         : No Position embedding on Special tokens
#
#     Normalization:
#     - rms         : RMSNorm (instead of LayerNorm)
#     - pn          : Pre-Norm (layer norm before the encoder) - implies norm eps of 1e-5
#     - npn         : No Post Norm (disables post-normalization layer)
#     - qkn         : QK Norm
#
#     Feed-Forward Network:
#     - swiglu      : SwiGLU FFN layer type (instead of standard FFN)
#
#     Activation:
#     - quick_gelu  : QuickGELU activation type
#     - ...
#
#     Regularization:
#     - ls          : Layer Scaling applied
#
#     Pooling/Reduction:
#     - avg         : Average pooling for sequence reduction
#     - ap          : Attention Pooling for sequence reduction
#     - aps         : Attention Pooling inc. Special tokens for sequence reduction
#
#     Custom Variants:
#     - c{N}        : Custom variant (N = version number) for models with fine-grained or non-standard
#                     modifications not fully reflected in the name


def register_rope_vit_configs(rope_vit: type[BaseNet]) -> None:
    registry.register_model_config(
        "rope_vit_t32",
        rope_vit,
        config={"patch_size": 32, **TINY},
    )
    registry.register_model_config(
        "rope_vit_t16",
        rope_vit,
        config={"patch_size": 16, **TINY},
    )
    registry.register_model_config(
        "rope_vit_t14",
        rope_vit,
        config={"patch_size": 14, **TINY},
    )
    registry.register_model_config(
        "rope_vit_s32",
        rope_vit,
        config={"patch_size": 32, **SMALL},
    )
    registry.register_model_config(
        "rope_vit_s16",
        rope_vit,
        config={"patch_size": 16, **SMALL},
    )
    registry.register_model_config(
        "rope_vit_s16_avg",
        rope_vit,
        config={"patch_size": 16, **SMALL, "class_token": False},
    )
    registry.register_model_config(
        "rope_i_vit_s16_pn_aps_c1",  # For PE Core - https://arxiv.org/abs/2504.13181
        rope_vit,
        config={
            "patch_size": 16,
            **SMALL,
            "pre_norm": True,
            "attn_pool_head": True,
            "attn_pool_num_heads": 8,
            "attn_pool_special_tokens": True,
            "norm_layer_eps": 1e-5,
            "rope_rot_type": "interleaved",
            "rope_grid_indexing": "xy",
            "rope_grid_offset": 1,
            "rope_temperature": 10000.0,
        },
    )
    registry.register_model_config(
        "rope_vit_s14",
        rope_vit,
        config={"patch_size": 14, **SMALL},
    )
    registry.register_model_config(
        "rope_vit_m32",
        rope_vit,
        config={"patch_size": 32, **MEDIUM},
    )
    registry.register_model_config(
        "rope_vit_m16",
        rope_vit,
        config={"patch_size": 16, **MEDIUM},
    )
    registry.register_model_config(
        "rope_vit_m14",
        rope_vit,
        config={"patch_size": 14, **MEDIUM},
    )
    registry.register_model_config(
        "rope_vit_b32",
        rope_vit,
        config={"patch_size": 32, **BASE, "drop_path_rate": 0.0},  # Override the BASE definition
    )
    registry.register_model_config(
        "rope_vit_b16",
        rope_vit,
        config={"patch_size": 16, **BASE},
    )
    registry.register_model_config(
        "rope_vit_b16_qkn_ls",
        rope_vit,
        config={"patch_size": 16, **BASE, "layer_scale_init_value": 1e-5, "qk_norm": True},
    )
    registry.register_model_config(
        "rope_i_vit_b16_pn_aps_c1",  # For PE Core - https://arxiv.org/abs/2504.13181
        rope_vit,
        config={
            "patch_size": 16,
            **BASE,
            "pre_norm": True,
            "attn_pool_head": True,
            "attn_pool_num_heads": 8,
            "attn_pool_special_tokens": True,
            "norm_layer_eps": 1e-5,
            "rope_rot_type": "interleaved",
            "rope_grid_indexing": "xy",
            "rope_grid_offset": 1,
            "rope_temperature": 10000.0,
        },
    )
    registry.register_model_config(
        "rope_vit_b14",
        rope_vit,
        config={"patch_size": 14, **BASE},
    )
    registry.register_model_config(
        "rope_vit_so150m_p14_ap",
        rope_vit,
        config={"patch_size": 14, **SO150, "class_token": False, "attn_pool_head": True},
    )
    registry.register_model_config(
        "rope_vit_l32",
        rope_vit,
        config={"patch_size": 32, **LARGE},
    )
    registry.register_model_config(
        "rope_vit_l16",
        rope_vit,
        config={"patch_size": 16, **LARGE},
    )
    registry.register_model_config(
        "rope_vit_l14",
        rope_vit,
        config={"patch_size": 14, **LARGE},
    )
    registry.register_model_config(
        "rope_i_vit_l14_pn_aps_c1",  # For PE Core - https://arxiv.org/abs/2504.13181
        rope_vit,
        config={
            "patch_size": 14,
            **LARGE,
            "pre_norm": True,
            "attn_pool_head": True,
            "attn_pool_num_heads": 8,
            "attn_pool_special_tokens": True,
            "norm_layer_eps": 1e-5,
            "rope_rot_type": "interleaved",
            "rope_grid_indexing": "xy",
            "rope_grid_offset": 1,
            "rope_temperature": 10000.0,
        },
    )
    registry.register_model_config(
        "rope_vit_so400m_p14_ap",
        rope_vit,
        config={"patch_size": 14, **SO400, "class_token": False, "attn_pool_head": True},
    )
    registry.register_model_config(
        "rope_vit_h16",
        rope_vit,
        config={"patch_size": 16, **HUGE},
    )
    registry.register_model_config(
        "rope_vit_h14",
        rope_vit,
        config={"patch_size": 14, **HUGE},
    )
    registry.register_model_config(  # From "Scaling Vision Transformers"
        "rope_vit_g16",
        rope_vit,
        config={"patch_size": 16, **GIANT},
    )
    registry.register_model_config(  # From "Scaling Vision Transformers"
        "rope_vit_g14",
        rope_vit,
        config={"patch_size": 14, **GIANT},
    )
    registry.register_model_config(  # From "Scaling Vision Transformers"
        "rope_vit_gigantic14",
        rope_vit,
        config={"patch_size": 14, **GIGANTIC},
    )

    # With registers
    ####################

    registry.register_model_config(
        "rope_vit_reg1_t32",
        rope_vit,
        config={"patch_size": 32, **TINY, "num_reg_tokens": 1},
    )
    registry.register_model_config(
        "rope_vit_reg1_t16",
        rope_vit,
        config={"patch_size": 16, **TINY, "num_reg_tokens": 1},
    )
    registry.register_model_config(
        "rope_vit_reg1_t14",
        rope_vit,
        config={"patch_size": 14, **TINY, "num_reg_tokens": 1},
    )
    registry.register_model_config(
        "rope_vit_reg1_s32",
        rope_vit,
        config={"patch_size": 32, **SMALL, "num_reg_tokens": 1},
    )
    registry.register_model_config(
        "rope_vit_reg1_s16",
        rope_vit,
        config={"patch_size": 16, **SMALL, "num_reg_tokens": 1},
    )
    registry.register_model_config(
        "rope_i_vit_reg1_s16_pn_npn_avg_c1",  # For PE Spatial - https://arxiv.org/abs/2504.13181
        rope_vit,
        config={
            "patch_size": 16,
            **SMALL,
            "num_reg_tokens": 1,
            "class_token": False,
            "pre_norm": True,
            "post_norm": False,
            "norm_layer_eps": 1e-5,
            "rope_rot_type": "interleaved",
            "rope_grid_indexing": "xy",
            "rope_grid_offset": 1,
            "rope_temperature": 10000.0,
        },
    )
    registry.register_model_config(
        "rope_vit_reg1_s14",
        rope_vit,
        config={"patch_size": 14, **SMALL, "num_reg_tokens": 1},
    )
    registry.register_model_config(
        "rope_vit_reg4_m32",
        rope_vit,
        config={"patch_size": 32, **MEDIUM, "num_reg_tokens": 4},
    )
    registry.register_model_config(
        "rope_vit_reg4_m16",
        rope_vit,
        config={"patch_size": 16, **MEDIUM, "num_reg_tokens": 4},
    )
    registry.register_model_config(
        "rope_vit_reg4_m16_rms_avg",
        rope_vit,
        config={"patch_size": 16, **MEDIUM, "num_reg_tokens": 4, "class_token": False, "norm_layer_type": "RMSNorm"},
    )
    registry.register_model_config(
        "rope_vit_reg4_m14",
        rope_vit,
        config={"patch_size": 14, **MEDIUM, "num_reg_tokens": 4},
    )
    registry.register_model_config(
        "rope_vit_reg4_m14_avg",
        rope_vit,
        config={"patch_size": 14, **MEDIUM, "num_reg_tokens": 4, "class_token": False},
    )
    registry.register_model_config(
        "rope_vit_reg4_b32",
        rope_vit,
        config={"patch_size": 32, **BASE, "num_reg_tokens": 4, "drop_path_rate": 0.0},  # Override the BASE definition
    )
    registry.register_model_config(
        "rope_vit_reg4_b16",
        rope_vit,
        config={"patch_size": 16, **BASE, "num_reg_tokens": 4},
    )
    registry.register_model_config(
        "rope_vit_reg4_b14",
        rope_vit,
        config={"patch_size": 14, **BASE, "num_reg_tokens": 4},
    )
    registry.register_model_config(
        "rope_vit_reg8_nps_b14_ap",
        rope_vit,
        config={
            "pos_embed_special_tokens": False,
            "patch_size": 14,
            **BASE,
            "num_reg_tokens": 8,
            "class_token": False,
            "attn_pool_head": True,
        },
    )
    registry.register_model_config(
        "rope_vit_reg4_so150m_p14_ap",
        rope_vit,
        config={"patch_size": 14, **SO150, "num_reg_tokens": 4, "class_token": False, "attn_pool_head": True},
    )
    registry.register_model_config(
        "rope_vit_reg8_so150m_p14_ap",
        rope_vit,
        config={"patch_size": 14, **SO150, "num_reg_tokens": 8, "class_token": False, "attn_pool_head": True},
    )
    registry.register_model_config(
        "rope_vit_reg8_so150m_p14_swiglu_rms_avg",
        rope_vit,
        config={
            "patch_size": 14,
            **SO150,
            "num_reg_tokens": 8,
            "class_token": False,
            "norm_layer_type": "RMSNorm",
            "mlp_layer_type": "SwiGLU_FFN",
        },
    )
    registry.register_model_config(
        "rope_vit_reg8_so150m_p14_swiglu_rms_ap",
        rope_vit,
        config={
            "patch_size": 14,
            **SO150,
            "num_reg_tokens": 8,
            "class_token": False,
            "attn_pool_head": True,
            "norm_layer_type": "RMSNorm",
            "mlp_layer_type": "SwiGLU_FFN",
        },
    )
    registry.register_model_config(
        "rope_vit_reg8_so150m_p14_swiglu_rms_aps",
        rope_vit,
        config={
            "patch_size": 14,
            **SO150,
            "num_reg_tokens": 8,
            "class_token": False,
            "attn_pool_head": True,
            "attn_pool_special_tokens": True,
            "norm_layer_type": "RMSNorm",
            "mlp_layer_type": "SwiGLU_FFN",
        },
    )
    registry.register_model_config(
        "rope_vit_reg4_l32",
        rope_vit,
        config={"patch_size": 32, **LARGE, "num_reg_tokens": 4},
    )
    registry.register_model_config(
        "rope_vit_reg4_l16",
        rope_vit,
        config={"patch_size": 16, **LARGE, "num_reg_tokens": 4},
    )
    registry.register_model_config(
        "rope_vit_reg4_l14",
        rope_vit,
        config={"patch_size": 14, **LARGE, "num_reg_tokens": 4},
    )
    registry.register_model_config(
        "rope_vit_reg8_l14_rms_ap",
        rope_vit,
        config={
            "patch_size": 14,
            **LARGE,
            "num_reg_tokens": 8,
            "class_token": False,
            "attn_pool_head": True,
            "norm_layer_type": "RMSNorm",
        },
    )
    registry.register_model_config(
        "rope_vit_reg8_so400m_p14_ap",
        rope_vit,
        config={"patch_size": 14, **SO400, "num_reg_tokens": 8, "class_token": False, "attn_pool_head": True},
    )
    registry.register_model_config(
        "rope_vit_reg4_h16",
        rope_vit,
        config={"patch_size": 16, **HUGE, "num_reg_tokens": 4},
    )
    registry.register_model_config(
        "rope_vit_reg4_h14",
        rope_vit,
        config={"patch_size": 14, **HUGE, "num_reg_tokens": 4},
    )
    registry.register_model_config(  # From "Scaling Vision Transformers"
        "rope_vit_reg4_g14",
        rope_vit,
        config={"patch_size": 14, **GIANT, "num_reg_tokens": 4},
    )



from birder.net.detection.deformable_detr import Deformable_DETR
from birder.net.detection.detr import DETR
from birder.net.detection.efficientdet import EfficientDet
from birder.net.detection.faster_rcnn import Faster_RCNN
from birder.net.detection.fcos import FCOS
from birder.net.detection.lw_detr import LW_DETR
from birder.net.detection.plain_detr import Plain_DETR
from birder.net.detection.retinanet import RetinaNet
from birder.net.detection.rt_detr_v1 import RT_DETR_v1
from birder.net.detection.rt_detr_v2 import RT_DETR_v2
from birder.net.detection.ssd import SSD
from birder.net.detection.ssdlite import SSDLite
from birder.net.detection.vitdet import ViTDet
from birder.net.detection.yolo_v2 import YOLO_v2
from birder.net.detection.yolo_v3 import YOLO_v3
from birder.net.detection.yolo_v4 import YOLO_v4
from birder.net.detection.yolo_v4_tiny import YOLO_v4_Tiny

__all__ = [
    "Deformable_DETR",
    "DETR",
    "EfficientDet",
    "Faster_RCNN",
    "FCOS",
    "LW_DETR",
    "Plain_DETR",
    "RetinaNet",
    "RT_DETR_v1",
    "RT_DETR_v2",
    "SSD",
    "SSDLite",
    "ViTDet",
    "YOLO_v2",
    "YOLO_v3",
    "YOLO_v4",
    "YOLO_v4_Tiny",
]



"""
LW-DETR (Lightweight DETR), adapted from
https://github.com/Atten4Vis/LW-DETR

Paper "LW-DETR: A Transformer Replacement to YOLO for Real-Time Detection", https://arxiv.org/abs/2406.03459

Changes from original:
* Uses lite_refpoint_refine mode only (fixed reference points across decoder layers)
* Uses bbox_reparam mode only (anchor-delta parameterization)

Generated by gpt-5.2-codex xhigh.
"""

# Reference license: Apache-2.0

import copy
import math
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import boxes as box_ops

from birder.common import training_utils
from birder.layers import LayerNorm2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.detection.base import DetectionBaseNet
from birder.net.detection.deformable_detr import HungarianMatcher
from birder.net.detection.deformable_detr import MultiScaleDeformableAttention
from birder.ops.soft_nms import SoftNMS


def _get_clones(module: nn.Module, N: int) -> nn.ModuleList:
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


def _apply_reference_delta(reference_points: torch.Tensor, bbox_delta: torch.Tensor) -> torch.Tensor:
    ref_xy = bbox_delta[..., :2] * reference_points[..., 2:] + reference_points[..., :2]
    ref_wh = bbox_delta[..., 2:].exp() * reference_points[..., 2:]
    return torch.concat((ref_xy, ref_wh), dim=-1)


def _gen_sine_embed_for_position(pos_tensor: torch.Tensor, dim: int, temperature: float = 10000.0) -> torch.Tensor:
    scale = 2 * math.pi
    dim_t = torch.arange(dim, dtype=torch.float32, device=pos_tensor.device)
    dim_t = temperature ** (2 * (dim_t // 2) / dim)

    x_embed = pos_tensor[..., 0] * scale
    y_embed = pos_tensor[..., 1] * scale
    pos_x = x_embed[..., None] / dim_t
    pos_y = y_embed[..., None] / dim_t
    pos_x = torch.stack((pos_x[..., 0::2].sin(), pos_x[..., 1::2].cos()), dim=-1).flatten(-2)
    pos_y = torch.stack((pos_y[..., 0::2].sin(), pos_y[..., 1::2].cos()), dim=-1).flatten(-2)

    if pos_tensor.size(-1) == 2:
        pos = torch.concat((pos_y, pos_x), dim=-1)
    elif pos_tensor.size(-1) == 4:
        w_embed = pos_tensor[..., 2] * scale
        h_embed = pos_tensor[..., 3] * scale
        pos_w = w_embed[..., None] / dim_t
        pos_h = h_embed[..., None] / dim_t
        pos_w = torch.stack((pos_w[..., 0::2].sin(), pos_w[..., 1::2].cos()), dim=-1).flatten(-2)
        pos_h = torch.stack((pos_h[..., 0::2].sin(), pos_h[..., 1::2].cos()), dim=-1).flatten(-2)
        pos = torch.concat((pos_y, pos_x, pos_w, pos_h), dim=-1)
    else:
        raise ValueError(f"Unknown pos_tensor shape(-1): {pos_tensor.size(-1)}")

    return pos


class MultiheadAttention(nn.Module):
    def __init__(self, d_model: int, num_heads: int, attn_drop: float = 0.0, proj_drop: float = 0.0) -> None:
        super().__init__()
        assert d_model % num_heads == 0, "d_model should be divisible by num_heads"

        self.num_heads = num_heads
        self.head_dim = d_model // num_heads
        self.scale = self.head_dim**-0.5

        self.q_proj = nn.Linear(d_model, d_model)
        self.k_proj = nn.Linear(d_model, d_model)
        self.v_proj = nn.Linear(d_model, d_model)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(d_model, d_model)
        self.proj_drop = nn.Dropout(proj_drop)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.xavier_uniform_(self.q_proj.weight)
        nn.init.xavier_uniform_(self.k_proj.weight)
        nn.init.xavier_uniform_(self.v_proj.weight)
        nn.init.xavier_uniform_(self.proj.weight)
        if self.q_proj.bias is not None:
            nn.init.zeros_(self.q_proj.bias)
            nn.init.zeros_(self.k_proj.bias)
            nn.init.zeros_(self.v_proj.bias)
            nn.init.zeros_(self.proj.bias)

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        key_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        B, l_q, C = query.size()
        q = self.q_proj(query).reshape(B, l_q, self.num_heads, self.head_dim).transpose(1, 2)
        k = self.k_proj(key).reshape(B, key.size(1), self.num_heads, self.head_dim).transpose(1, 2)
        v = self.v_proj(value).reshape(B, value.size(1), self.num_heads, self.head_dim).transpose(1, 2)

        if key_padding_mask is not None:
            # key_padding_mask is expected to be boolean (True = masked)
            # SDPA expects True = attend, so we invert
            attn_mask = ~key_padding_mask[:, None, None, :]
        else:
            attn_mask = None

        attn = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, attn_mask=attn_mask, dropout_p=self.attn_drop.p if self.training else 0.0, scale=self.scale
        )

        attn = attn.transpose(1, 2).reshape(B, l_q, C)
        x = self.proj(attn)
        x = self.proj_drop(x)

        return x


class Bottleneck(nn.Module):
    def __init__(self, channels: int, expansion: float = 1.0) -> None:
        super().__init__()
        hidden = int(channels * expansion)
        self.conv1 = Conv2dNormActivation(
            channels, hidden, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), activation_layer=nn.SiLU
        )
        self.conv2 = Conv2dNormActivation(
            hidden, channels, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), activation_layer=nn.SiLU
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x + self.conv2(self.conv1(x))


class C2f(nn.Module):
    """
    CSP Bottleneck with 2 convolutions
    """

    def __init__(self, in_channels: int, out_channels: int, num_blocks: int, expansion: float = 0.5) -> None:
        super().__init__()
        hidden = int(out_channels * expansion)
        self.conv1 = Conv2dNormActivation(
            in_channels, hidden * 2, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=nn.SiLU
        )
        self.blocks = nn.ModuleList([Bottleneck(hidden, expansion=1.0) for _ in range(num_blocks)])
        self.conv2 = Conv2dNormActivation(
            hidden * (2 + num_blocks),
            out_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            activation_layer=nn.SiLU,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = self.conv1(x)
        c = y.size(1) // 2
        y = list(y.split([c, c], dim=1))
        y.extend(block(y[-1]) for block in self.blocks)

        return self.conv2(torch.concat(y, dim=1))


class MultiScaleProjector(nn.Module):
    """
    LW-DETR Multi-Scale Projector that creates pyramid features from backbone feature maps

    For each output scale, applies per-level sampling (upsample/downsample) to each input
    feature map, resizes to match, concatenates them, and passes through C2f + LayerNorm.
    """

    def __init__(self, in_channels: list[int], out_channels: int, scale_factors: list[float], num_blocks: int) -> None:
        super().__init__()
        self.use_extra_pool = False

        self.stage_scales: list[float] = []
        self.stages = nn.ModuleList()
        for scale in scale_factors:
            scale_samplers: list[nn.Module] = []
            fused_channels = 0

            for in_dim in in_channels:
                out_dim = in_dim
                if scale == 4.0:
                    # 4x upsample via two ConvTranspose2d
                    layers: nn.Module = nn.Sequential(
                        nn.ConvTranspose2d(in_dim, in_dim // 2, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)),
                        LayerNorm2d(in_dim // 2, eps=1e-6),
                        nn.GELU(),
                        nn.ConvTranspose2d(in_dim // 2, in_dim // 4, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)),
                    )
                    out_dim = in_dim // 4
                elif scale == 2.0:
                    # 2x upsample
                    if in_dim > 512:
                        layers = nn.Sequential(
                            Conv2dNormActivation(in_dim, in_dim // 2, kernel_size=1, activation_layer=nn.ReLU),
                            nn.ConvTranspose2d(
                                in_dim // 2, in_dim // 4, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)
                            ),
                        )
                        out_dim = in_dim // 4
                    else:
                        layers = nn.ConvTranspose2d(
                            in_dim, in_dim // 2, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)
                        )
                        out_dim = in_dim // 2
                elif scale == 1.0:
                    # Identity (will resize to match in forward)
                    layers = nn.Identity()
                    out_dim = in_dim
                elif scale == 0.5:
                    # 2x downsample
                    layers = Conv2dNormActivation(
                        in_dim, in_dim, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), activation_layer=nn.SiLU
                    )
                    out_dim = in_dim
                elif scale == 0.25:
                    # Will use max pooling on previous scale output
                    self.use_extra_pool = True
                    continue
                else:
                    raise ValueError(f"Unsupported scale_factor: {scale}")

                scale_samplers.append(layers)
                fused_channels += out_dim

            if len(scale_samplers) > 0:
                fuser = nn.Sequential(
                    C2f(fused_channels, out_channels, num_blocks),
                    LayerNorm2d(out_channels, eps=1e-6),
                )
                stage = nn.ModuleDict(
                    {
                        "samplers": nn.ModuleList(scale_samplers),
                        "fuser": fuser,
                    }
                )
                self.stage_scales.append(float(scale))
                self.stages.append(stage)

    def forward(self, features: list[torch.Tensor]) -> list[torch.Tensor]:
        results: list[torch.Tensor] = []

        # Use the middle feature level as reference for target size computation
        ref_idx = len(features) // 2
        ref_h, ref_w = features[ref_idx].shape[-2:]

        for stage_idx, stage in enumerate(self.stages):
            scale = self.stage_scales[stage_idx]
            target_h = int(ref_h * scale)
            target_w = int(ref_w * scale)

            samplers = stage["samplers"]
            fuser = stage["fuser"]

            fused_feats: list[torch.Tensor] = []
            for level_idx, sampler in enumerate(samplers):
                feat = sampler(features[level_idx])
                if feat.shape[-2:] != (target_h, target_w):
                    feat = F.interpolate(feat, size=(target_h, target_w), mode="nearest")

                fused_feats.append(feat)

            fused = torch.concat(fused_feats, dim=1)
            results.append(fuser(fused))

        if self.use_extra_pool is True:
            results.append(F.max_pool2d(results[-1], kernel_size=(1, 1), stride=(2, 2), padding=(0, 0)))

        return results


class LWDeformableTransformerDecoderLayer(nn.Module):
    def __init__(
        self,
        d_model: int,
        d_ffn: int,
        dropout: float,
        n_levels: int,
        self_heads: int,
        cross_heads: int,
        n_points: int,
        num_queries: int,
        group_detr: int,
    ) -> None:
        super().__init__()
        self.num_queries = num_queries
        self.group_detr = group_detr

        self.self_attn = MultiheadAttention(d_model, self_heads, attn_drop=dropout)
        self.norm1 = nn.LayerNorm(d_model)

        self.cross_attn = MultiScaleDeformableAttention(d_model, n_levels, cross_heads, n_points)
        self.norm2 = nn.LayerNorm(d_model)

        self.linear1 = nn.Linear(d_model, d_ffn)
        self.linear2 = nn.Linear(d_ffn, d_model)
        self.norm3 = nn.LayerNorm(d_model)

        self.activation = nn.ReLU()
        self.dropout = nn.Dropout(dropout)

    def forward(
        self,
        tgt: torch.Tensor,
        query_pos: torch.Tensor,
        reference_points: torch.Tensor,
        src: torch.Tensor,
        src_spatial_shapes: torch.Tensor,
        src_shapes: list[list[int]],
        level_start_index: torch.Tensor,
        src_padding_mask: Optional[torch.Tensor],
    ) -> torch.Tensor:
        q_k = tgt + query_pos

        # Group-DETR: during training, split queries by groups so each group attends only within itself
        if self.training is True and self.group_detr > 1:
            B, N, C = q_k.size()
            num_queries_per_group = N // self.group_detr
            qk_grouped = q_k.view(B, self.group_detr, num_queries_per_group, C)
            qk_grouped = qk_grouped.reshape(B * self.group_detr, num_queries_per_group, C)

            tgt_grouped = tgt.view(B, self.group_detr, num_queries_per_group, C)
            tgt_grouped = tgt_grouped.reshape(B * self.group_detr, num_queries_per_group, C)

            tgt2 = self.self_attn(qk_grouped, qk_grouped, value=tgt_grouped)
            tgt2 = tgt2.view(B, self.group_detr, num_queries_per_group, C)
            tgt2 = tgt2.reshape(B, N, C)
        else:
            tgt2 = self.self_attn(q_k, q_k, value=tgt)

        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm1(tgt)

        tgt2 = self.cross_attn(
            tgt + query_pos, reference_points, src, src_spatial_shapes, level_start_index, src_padding_mask, src_shapes
        )
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm2(tgt)

        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt))))
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm3(tgt)

        return tgt


# pylint: disable=invalid-name
class LW_DETRDecoder(nn.Module):
    """
    LW-DETR Decoder with learned queries and fixed reference points

    Uses learned content queries and reference point embeddings. Supports Group-DETR
    training where queries are replicated into multiple groups with independent matching.
    In two-stage mode, reference points come from encoder proposals instead of learned embeddings.
    """

    def __init__(
        self,
        hidden_dim: int,
        num_classes: int,
        num_queries: int,
        num_decoder_layers: int,
        num_levels: int,
        num_self_heads: int,
        num_cross_heads: int,
        dim_feedforward: int,
        dropout: float,
        num_decoder_points: int,
        group_detr: int,
        two_stage: bool,
    ) -> None:
        super().__init__()
        self.hidden_dim = hidden_dim
        self.num_queries = num_queries
        self.num_layers = num_decoder_layers
        self.num_levels = num_levels
        self.group_detr = group_detr
        self.two_stage = two_stage

        # Learnable content queries and reference points (per Group-DETR)
        self.content_queries = nn.Parameter(torch.empty(num_queries * group_detr, hidden_dim))
        self.reference_point_embed = nn.Parameter(torch.empty(num_queries * group_detr, 4))

        decoder_layer = LWDeformableTransformerDecoderLayer(
            hidden_dim,
            dim_feedforward,
            dropout,
            num_levels,
            num_self_heads,
            num_cross_heads,
            num_decoder_points,
            num_queries,
            group_detr,
        )
        self.layers = _get_clones(decoder_layer, num_decoder_layers)

        # Sinusoidal embeddings: 4D coords -> 4 * (hidden_dim/2) = 2 * hidden_dim
        self.query_pos_head = MLP(2 * hidden_dim, [hidden_dim, hidden_dim], activation_layer=nn.ReLU)

        self.class_embed = nn.Linear(hidden_dim, num_classes)
        self.bbox_embed = MLP(hidden_dim, [hidden_dim, hidden_dim, 4], activation_layer=nn.ReLU)

        # Weights initialization
        prior_prob = 0.01
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        nn.init.xavier_uniform_(self.content_queries)
        nn.init.zeros_(self.reference_point_embed)
        nn.init.constant_(self.class_embed.bias, bias_value)
        nn.init.zeros_(self.bbox_embed[-2].weight)
        nn.init.zeros_(self.bbox_embed[-2].bias)

    def reset_classifier(self, num_classes: int) -> None:
        self.class_embed = nn.Linear(self.hidden_dim, num_classes)
        prior_prob = 0.01
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        nn.init.constant_(self.class_embed.bias, bias_value)

    # pylint: disable=too-many-locals,too-many-branches
    def forward(
        self,
        feats: list[torch.Tensor],
        spatial_shapes: list[list[int]],
        level_start_index: list[int],
        valid_ratios: torch.Tensor,
        padding_mask: Optional[list[torch.Tensor]] = None,
        enc_reference_points: Optional[torch.Tensor] = None,
        return_intermediates: bool = True,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        memory = []
        mask_flatten = []
        for idx, feat in enumerate(feats):
            feat = feat.flatten(2).permute(0, 2, 1)
            memory.append(feat)
            if padding_mask is not None:
                mask_flatten.append(padding_mask[idx].flatten(1))

        memory = torch.concat(memory, dim=1)
        memory_padding_mask = torch.concat(mask_flatten, dim=1) if len(mask_flatten) > 0 else None

        # Two-stage mode: combine learned reference point deltas with encoder proposals
        if self.two_stage is True:
            assert enc_reference_points is not None
            # Slice content queries and reference point embeddings to match encoder proposals
            num_enc_queries = enc_reference_points.size(1)
            if self.training is True:
                if self.group_detr > 1:
                    if num_enc_queries % self.group_detr != 0:
                        raise ValueError("num_enc_queries must be divisible by group_detr in training mode")
                    num_queries_per_group = num_enc_queries // self.group_detr
                    content_queries_weight = self.content_queries.reshape(
                        self.group_detr, self.num_queries, self.hidden_dim
                    )[:, :num_queries_per_group].reshape(-1, self.hidden_dim)
                    refpoint_embed_weight = self.reference_point_embed.reshape(self.group_detr, self.num_queries, 4)[
                        :, :num_queries_per_group
                    ].reshape(-1, 4)
                else:
                    content_queries_weight = self.content_queries[:num_enc_queries]
                    refpoint_embed_weight = self.reference_point_embed[:num_enc_queries]
            else:
                num_to_use = min(self.num_queries, num_enc_queries)
                content_queries_weight = self.content_queries[:num_to_use]
                refpoint_embed_weight = self.reference_point_embed[:num_to_use]
                enc_reference_points = enc_reference_points[:, :num_to_use]

            # Learned embeddings act as deltas on top of encoder proposals (bbox_reparam formula)
            refpoint_embed = refpoint_embed_weight.unsqueeze(0).expand(memory.size(0), -1, -1)
            refpoint_cxcy = refpoint_embed[..., :2] * enc_reference_points[..., 2:] + enc_reference_points[..., :2]
            refpoint_wh = refpoint_embed[..., 2:].exp() * enc_reference_points[..., 2:]
            reference_points = torch.concat([refpoint_cxcy, refpoint_wh], dim=-1)
        else:
            if self.training is True:
                content_queries_weight = self.content_queries
                refpoint_embed_weight = self.reference_point_embed
            else:
                content_queries_weight = self.content_queries[: self.num_queries]
                refpoint_embed_weight = self.reference_point_embed[: self.num_queries]

            init_ref_points_unact = refpoint_embed_weight.unsqueeze(0).expand(memory.size(0), -1, -1)

            # One-stage mode: reference points are un-activated, need sigmoid
            reference_points = init_ref_points_unact.sigmoid()

        target = content_queries_weight.unsqueeze(0).expand(memory.size(0), -1, -1)

        spatial_shapes_tensor = torch.tensor(spatial_shapes, dtype=torch.long, device=memory.device)
        level_start_index_tensor = torch.tensor(level_start_index, dtype=torch.long, device=memory.device)

        # lite_refpoint_refine mode: compute query_pos ONCE from initial reference points
        # Reference points stay fixed throughout all decoder layers
        if reference_points.shape[-1] == 4:
            reference_points_input = (
                reference_points[:, :, None] * torch.concat([valid_ratios, valid_ratios], dim=-1)[:, None]
            )
        else:
            reference_points_input = reference_points[:, :, None] * valid_ratios[:, None]

        query_sine_embed = _gen_sine_embed_for_position(reference_points_input[:, :, 0], self.hidden_dim // 2)
        query_pos = self.query_pos_head(query_sine_embed)

        bboxes_list: list[torch.Tensor] = []
        logits_list: list[torch.Tensor] = []
        layer_bboxes = torch.empty(0, device=target.device, dtype=target.dtype)
        class_logits = torch.empty(0, device=target.device, dtype=target.dtype)
        last_idx = len(self.layers) - 1
        for layer_idx, decoder_layer in enumerate(self.layers):
            target = decoder_layer(
                target,
                query_pos,
                reference_points_input,
                memory,
                spatial_shapes_tensor,
                spatial_shapes,
                level_start_index_tensor,
                memory_padding_mask,
            )

            if return_intermediates is True or layer_idx == last_idx:
                # Each layer predicts boxes from the same initial reference points
                bbox_delta = self.bbox_embed(target)
                layer_bboxes = _apply_reference_delta(reference_points, bbox_delta)
                class_logits = self.class_embed(target)
                if return_intermediates is True:
                    bboxes_list.append(layer_bboxes)
                    logits_list.append(class_logits)

        if return_intermediates is True:
            out_bboxes = torch.stack(bboxes_list)
            out_logits = torch.stack(logits_list)
        else:
            out_bboxes = layer_bboxes
            out_logits = class_logits

        return (out_bboxes, out_logits)


# pylint: disable=invalid-name
class LW_DETR(DetectionBaseNet):
    default_size = (640, 640)

    def __init__(
        self,
        num_classes: int,
        backbone: DetectorBackbone,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
        export_mode: bool = False,
    ) -> None:
        super().__init__(num_classes, backbone, config=config, size=size, export_mode=export_mode)
        assert self.config is not None, "must set config"

        # Sigmoid based classification (like multi-label networks)
        self.num_classes = self.num_classes - 1

        dropout = 0.0
        num_decoder_layers: int = self.config.get("num_decoder_layers", 3)
        projector_dim: int = self.config.get("projector_dim", 256)
        projector_scale_factors: list[float] = self.config["projector_scale_factors"]
        projector_num_blocks: int = self.config.get("projector_num_blocks", 3)
        decoder_dim: int = self.config.get("decoder_dim", 256)
        decoder_self_heads: int = self.config.get("decoder_self_heads", 8)
        decoder_cross_heads: int = self.config.get("decoder_cross_heads", decoder_self_heads)
        decoder_points: int = self.config.get("decoder_points", 2)
        num_queries: int = self.config.get("num_queries", 300)
        two_stage: bool = self.config.get("two_stage", False)
        soft_nms: bool = self.config.get("soft_nms", False)

        self.soft_nms = None
        if soft_nms is True:
            self.soft_nms = SoftNMS()

        self.ia_alpha = 0.25
        self.loss_class = 1.0
        self.loss_bbox = 5.0
        self.loss_giou = 2.0
        self.aux_loss = True
        self.num_queries = num_queries
        self.two_stage = two_stage
        self.group_detr = 13

        self.backbone.return_channels = self.backbone.return_channels[-3:]
        self.backbone.return_stages = self.backbone.return_stages[-3:]

        num_decoder_levels = len(projector_scale_factors)

        self.projector = MultiScaleProjector(
            in_channels=list(self.backbone.return_channels),
            out_channels=projector_dim,
            scale_factors=projector_scale_factors,
            num_blocks=projector_num_blocks,
        )
        self.decoder_input_proj = nn.ModuleList(
            [
                (
                    nn.Conv2d(projector_dim, decoder_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
                    if projector_dim != decoder_dim
                    else nn.Identity()
                )
                for _ in range(num_decoder_levels)
            ]
        )
        self.decoder = LW_DETRDecoder(
            hidden_dim=decoder_dim,
            num_classes=self.num_classes,
            num_queries=self.num_queries,
            num_decoder_layers=num_decoder_layers,
            num_levels=num_decoder_levels,
            num_self_heads=decoder_self_heads,
            num_cross_heads=decoder_cross_heads,
            dim_feedforward=decoder_dim * 4,
            dropout=dropout,
            num_decoder_points=decoder_points,
            group_detr=self.group_detr,
            two_stage=self.two_stage,
        )

        # Two-stage encoder output modules
        if self.two_stage is True:
            self.enc_output = nn.ModuleList([nn.Linear(decoder_dim, decoder_dim) for _ in range(self.group_detr)])
            self.enc_output_norm = nn.ModuleList([nn.LayerNorm(decoder_dim) for _ in range(self.group_detr)])
            # Encoder output class and bbox embeddings
            self.enc_out_class_embed = nn.ModuleList(
                [copy.deepcopy(self.decoder.class_embed) for _ in range(self.group_detr)]
            )
            self.enc_out_bbox_embed = nn.ModuleList(
                [copy.deepcopy(self.decoder.bbox_embed) for _ in range(self.group_detr)]
            )

            # Weights initialization
            for enc_output in self.enc_output:
                nn.init.xavier_uniform_(enc_output.weight)
                if enc_output.bias is not None:
                    nn.init.zeros_(enc_output.bias)

            for enc_output_norm in self.enc_output_norm:
                nn.init.ones_(enc_output_norm.weight)
                nn.init.zeros_(enc_output_norm.bias)

        else:
            self.enc_output = nn.ModuleList([nn.Identity()])
            self.enc_output_norm = nn.ModuleList([nn.Identity()])
            self.enc_out_class_embed = nn.ModuleList([nn.Identity()])
            self.enc_out_bbox_embed = nn.ModuleList([nn.Identity()])

        self.decoder_dim = decoder_dim
        self.matcher = HungarianMatcher(cost_class=self.loss_class, cost_bbox=self.loss_bbox, cost_giou=self.loss_giou)

        if self.export_mode is False:
            self.forward = torch.compiler.disable(recursive=False)(self.forward)  # type: ignore[method-assign]

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        self.decoder.reset_classifier(num_classes)
        if self.two_stage is True:
            self.enc_out_class_embed = nn.ModuleList(
                [copy.deepcopy(self.decoder.class_embed) for _ in range(self.group_detr)]
            )

    def freeze(self, freeze_classifier: bool = True) -> None:
        for param in self.parameters():
            param.requires_grad_(False)
        if freeze_classifier is False:
            for param in self.decoder.class_embed.parameters():
                param.requires_grad_(True)

            if self.two_stage is True:
                for enc_class_embed in self.enc_out_class_embed:
                    for param in enc_class_embed.parameters():
                        param.requires_grad_(True)

    @staticmethod
    def _get_src_permutation_idx(indices: list[tuple[torch.Tensor, torch.Tensor]]) -> tuple[torch.Tensor, torch.Tensor]:
        batch_idx = torch.concat([torch.full_like(src, i) for i, (src, _) in enumerate(indices)])
        src_idx = torch.concat([src for (src, _) in indices])
        return (batch_idx, src_idx)

    def _class_loss(
        self,
        cls_logits: torch.Tensor,
        box_output: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        indices: list[tuple[torch.Tensor, torch.Tensor]],
        num_boxes: float,
    ) -> torch.Tensor:
        idx = self._get_src_permutation_idx(indices)
        pred_scores = cls_logits.sigmoid()
        target_scores = torch.zeros_like(cls_logits)
        pos_mask = torch.zeros_like(cls_logits, dtype=torch.bool)
        if idx[0].numel() > 0:
            target_classes = torch.concat([t["labels"][J] for t, (_, J) in zip(targets, indices)], dim=0)
            src_boxes = box_output[idx]
            target_boxes = torch.concat([t["boxes"][J] for t, (_, J) in zip(targets, indices)], dim=0)
            iou = box_ops.box_iou(
                box_ops.box_convert(src_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
                box_ops.box_convert(target_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
            ).diag()
            iou = iou.clamp(min=0).detach()
            mixed_target = pred_scores.detach()[idx[0], idx[1], target_classes] ** self.ia_alpha
            mixed_target = mixed_target * (iou ** (1.0 - self.ia_alpha))
            target_scores[idx[0], idx[1], target_classes] = mixed_target.to(target_scores.dtype)
            pos_mask[idx[0], idx[1], target_classes] = True

        neg_mask = ~pos_mask

        pos_loss = F.binary_cross_entropy_with_logits(cls_logits, target_scores, reduction="none")
        neg_loss = F.binary_cross_entropy_with_logits(cls_logits, torch.zeros_like(cls_logits), reduction="none")
        neg_loss = neg_loss * pred_scores.pow(2)

        loss_pos = pos_loss[pos_mask].sum()
        loss_neg = neg_loss[neg_mask].sum()
        return (loss_pos + loss_neg) / num_boxes

    def _box_loss(
        self,
        box_output: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        indices: list[tuple[torch.Tensor, torch.Tensor]],
        num_boxes: float,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        idx = self._get_src_permutation_idx(indices)
        src_boxes = box_output[idx]
        target_boxes = torch.concat([t["boxes"][i] for t, (_, i) in zip(targets, indices)], dim=0)

        loss_bbox = F.l1_loss(src_boxes, target_boxes, reduction="none")
        loss_bbox = loss_bbox.sum() / num_boxes

        loss_giou = 1 - torch.diag(
            box_ops.generalized_box_iou(
                box_ops.box_convert(src_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
                box_ops.box_convert(target_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
            )
        )
        loss_giou = loss_giou.sum() / num_boxes

        return (loss_bbox, loss_giou)

    def _match_with_groups(
        self, cls_logits: torch.Tensor, box_output: torch.Tensor, targets: list[dict[str, torch.Tensor]]
    ) -> list[tuple[torch.Tensor, torch.Tensor]]:
        """
        Perform Hungarian matching with Group-DETR support

        Reshapes queries into a larger batch (B * G) so a single matcher call
        can be used, then combines indices with proper group offsets.
        """

        if self.group_detr <= 1:
            return self.matcher(cls_logits, box_output, targets)  # type: ignore[no-any-return]

        batch_size = cls_logits.size(0)
        group_detr = self.group_detr
        total_queries = cls_logits.size(1)
        num_queries_per_group = total_queries // group_detr

        # Reshape from (B, G*Q, C) to (B*G, Q, C) to batch all groups together
        cls_logits_grouped = cls_logits.reshape(batch_size * group_detr, num_queries_per_group, -1)
        box_output_grouped = box_output.reshape(batch_size * group_detr, num_queries_per_group, -1)

        # Duplicate targets for each group: each group within a batch matches against the same targets
        targets_grouped = [targets[batch_idx] for batch_idx in range(batch_size) for _ in range(group_detr)]

        grouped_indices = self.matcher(cls_logits_grouped, box_output_grouped, targets_grouped)

        all_indices: list[tuple[torch.Tensor, torch.Tensor]] = []
        for batch_idx in range(batch_size):
            src_indices_list: list[torch.Tensor] = []
            tgt_indices_list: list[torch.Tensor] = []

            base = batch_idx * group_detr
            for group_idx in range(group_detr):
                src_idx, tgt_idx = grouped_indices[base + group_idx]
                # Adjust source indices by group offset to map back to original query positions
                src_indices_list.append(src_idx + group_idx * num_queries_per_group)
                tgt_indices_list.append(tgt_idx)

            # Combine indices from all groups for this batch element
            all_indices.append((torch.concat(src_indices_list), torch.concat(tgt_indices_list)))

        return all_indices

    def _compute_loss_from_outputs(
        self, targets: list[dict[str, torch.Tensor]], cls_logits: torch.Tensor, box_output: torch.Tensor
    ) -> dict[str, torch.Tensor]:
        num_boxes = sum(len(t["labels"]) for t in targets)
        num_boxes = torch.as_tensor(num_boxes, dtype=torch.float, device=cls_logits.device)
        if training_utils.is_dist_available_and_initialized() is True:
            torch.distributed.all_reduce(num_boxes)

        num_boxes = torch.clamp(num_boxes / training_utils.get_world_size(), min=1)

        # Normalize by group_detr for Group-DETR training
        num_boxes_normalized = num_boxes * self.group_detr

        loss_ce_list = []
        loss_bbox_list = []
        loss_giou_list = []
        for idx in range(cls_logits.size(0)):
            indices = self._match_with_groups(cls_logits[idx], box_output[idx], targets)

            loss_ce_list.append(
                self._class_loss(cls_logits[idx], box_output[idx], targets, indices, num_boxes_normalized)
            )
            loss_bbox_i, loss_giou_i = self._box_loss(box_output[idx], targets, indices, num_boxes_normalized)
            loss_bbox_list.append(loss_bbox_i)
            loss_giou_list.append(loss_giou_i)

        loss_ce = torch.stack(loss_ce_list).sum() * self.loss_class
        loss_bbox = torch.stack(loss_bbox_list).sum() * self.loss_bbox
        loss_giou = torch.stack(loss_giou_list).sum() * self.loss_giou

        losses = {"labels": loss_ce, "boxes": loss_bbox, "giou": loss_giou}

        return losses

    def _compute_enc_output_loss(
        self, targets: list[dict[str, torch.Tensor]], enc_cls_logits: torch.Tensor, enc_box_output: torch.Tensor
    ) -> dict[str, torch.Tensor]:
        num_boxes = sum(len(t["labels"]) for t in targets)
        num_boxes = torch.as_tensor(num_boxes, dtype=torch.float, device=enc_cls_logits.device)
        if training_utils.is_dist_available_and_initialized() is True:
            torch.distributed.all_reduce(num_boxes)

        num_boxes = torch.clamp(num_boxes / training_utils.get_world_size(), min=1)
        num_boxes_normalized = num_boxes * self.group_detr

        indices = self._match_with_groups(enc_cls_logits, enc_box_output, targets)
        loss_ce = self._class_loss(enc_cls_logits, enc_box_output, targets, indices, num_boxes_normalized)
        loss_bbox, loss_giou = self._box_loss(enc_box_output, targets, indices, num_boxes_normalized)

        return {
            "labels_enc": loss_ce * self.loss_class,
            "boxes_enc": loss_bbox * self.loss_bbox,
            "giou_enc": loss_giou * self.loss_giou,
        }

    @torch.jit.unused  # type: ignore[untyped-decorator]
    @torch.compiler.disable()  # type: ignore[untyped-decorator]
    def compute_loss(
        self,
        targets: list[dict[str, torch.Tensor]],
        out_logits: torch.Tensor,
        out_bboxes: torch.Tensor,
        image_sizes: torch.Tensor,
        enc_cls_logits: Optional[torch.Tensor] = None,
        enc_box_output: Optional[torch.Tensor] = None,
    ) -> dict[str, torch.Tensor]:
        losses: dict[str, torch.Tensor] = {}

        # Compute encoder output loss first (before modifying targets in-place)
        if self.two_stage is True and enc_cls_logits is not None and enc_box_output is not None:
            enc_targets = copy.deepcopy(targets)
            for idx, target in enumerate(enc_targets):
                boxes = target["boxes"]
                boxes = box_ops.box_convert(boxes, in_fmt="xyxy", out_fmt="cxcywh")
                scale = image_sizes[idx].flip(0).repeat(2).float()  # flip to [W, H], repeat to [W, H, W, H]
                boxes = boxes / scale
                enc_targets[idx]["boxes"] = boxes
                enc_targets[idx]["labels"] = target["labels"] - 1

            enc_losses = self._compute_enc_output_loss(enc_targets, enc_cls_logits, enc_box_output)
            losses.update(enc_losses)

        # Transform targets for decoder loss
        for idx, target in enumerate(targets):
            boxes = target["boxes"]
            boxes = box_ops.box_convert(boxes, in_fmt="xyxy", out_fmt="cxcywh")
            scale = image_sizes[idx].flip(0).repeat(2).float()  # flip to [W, H], repeat to [W, H, W, H]
            boxes = boxes / scale
            targets[idx]["boxes"] = boxes
            targets[idx]["labels"] = target["labels"] - 1

        if self.aux_loss is False:
            out_bboxes = out_bboxes[-1:].contiguous()
            out_logits = out_logits[-1:].contiguous()

        decoder_losses = self._compute_loss_from_outputs(targets, out_logits, out_bboxes)
        losses.update(decoder_losses)

        return losses

    @staticmethod
    def _get_valid_ratio(mask: torch.Tensor) -> torch.Tensor:
        _, H, W = mask.size()
        valid_h = torch.sum(~mask[:, :, 0], 1)
        valid_w = torch.sum(~mask[:, 0, :], 1)
        valid_ratio_h = valid_h.float() / H
        valid_ratio_w = valid_w.float() / W
        return torch.stack([valid_ratio_w, valid_ratio_h], dim=-1)

    def gen_encoder_output_proposals(
        self, memory: torch.Tensor, memory_padding_mask: Optional[torch.Tensor], spatial_shapes: list[list[int]]
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Generate encoder output proposals for two-stage detection

        Creates bounding box proposals from encoder memory by generating a grid of anchors
        across all spatial positions and feature levels.
        """

        N = memory.size(0)
        proposals = []
        cur = 0
        for lvl, (H, W) in enumerate(spatial_shapes):
            if memory_padding_mask is not None:
                mask_flatten = memory_padding_mask[:, cur : cur + H * W].view(N, H, W, 1)
                valid_h = torch.sum(~mask_flatten[:, :, 0, 0], 1)
                valid_w = torch.sum(~mask_flatten[:, 0, :, 0], 1)
            else:
                valid_h = torch.full((N,), H, dtype=torch.float32, device=memory.device)
                valid_w = torch.full((N,), W, dtype=torch.float32, device=memory.device)

            grid_y, grid_x = torch.meshgrid(
                torch.linspace(0, H - 1, H, dtype=torch.float32, device=memory.device),
                torch.linspace(0, W - 1, W, dtype=torch.float32, device=memory.device),
                indexing="ij",
            )
            grid = torch.stack([grid_x, grid_y], dim=-1)  # (H, W, 2)

            scale = torch.stack([valid_w, valid_h], dim=1).view(N, 1, 1, 2)
            grid = (grid.unsqueeze(0).expand(N, -1, -1, -1) + 0.5) / scale
            wh = torch.ones_like(grid) * 0.05 * (2.0**lvl)
            proposal = torch.concat([grid, wh], dim=-1).view(N, -1, 4)
            proposals.append(proposal)
            cur += H * W

        output_proposals = torch.concat(proposals, dim=1)
        output_proposals_valid = ((output_proposals > 0.01) & (output_proposals < 0.99)).all(dim=-1, keepdim=True)

        # bbox_reparam mode: keep proposals in normalized [0,1] space (no un-sigmoid transform)
        if memory_padding_mask is not None:
            output_proposals = output_proposals.masked_fill(memory_padding_mask.unsqueeze(-1), 0.0)

        output_proposals = output_proposals.masked_fill(~output_proposals_valid, 0.0)

        output_memory = memory
        if memory_padding_mask is not None:
            output_memory = output_memory.masked_fill(memory_padding_mask.unsqueeze(-1), 0.0)

        output_memory = output_memory.masked_fill(~output_proposals_valid, 0.0)

        return (output_memory, output_proposals.to(memory.dtype))

    def postprocess_detections(
        self, class_logits: torch.Tensor, box_regression: torch.Tensor, image_sizes: torch.Tensor
    ) -> list[dict[str, torch.Tensor]]:
        prob = class_logits.sigmoid()
        topk = min(self.num_queries, prob.shape[1] * prob.shape[2])
        topk_values, topk_indexes = torch.topk(prob.view(prob.shape[0], -1), k=topk, dim=1)
        scores = topk_values
        topk_boxes = topk_indexes // prob.shape[2]
        labels = topk_indexes % prob.shape[2]
        labels += 1  # background offset

        boxes = box_ops.box_convert(box_regression, in_fmt="cxcywh", out_fmt="xyxy")
        boxes = torch.gather(boxes, 1, topk_boxes.unsqueeze(-1).expand(-1, -1, 4))

        img_h, img_w = image_sizes.unbind(1)
        scale_fct = torch.stack([img_w, img_h, img_w, img_h], dim=1)
        boxes = boxes * scale_fct[:, None, :]

        detections: list[dict[str, torch.Tensor]] = []
        for s, lbl, b in zip(scores, labels, boxes):
            if self.soft_nms is not None:
                soft_scores, keep = self.soft_nms(b, s, lbl, score_threshold=0.001)
                s[keep] = soft_scores
                b = b[keep]
                s = s[keep]
                lbl = lbl[keep]

            detections.append({"boxes": b, "scores": s, "labels": lbl})

        return detections

    # pylint: disable=too-many-locals
    def forward_net(
        self, x: torch.Tensor, masks: Optional[torch.Tensor]
    ) -> tuple[Optional[torch.Tensor], Optional[torch.Tensor], torch.Tensor, torch.Tensor]:
        features: dict[str, torch.Tensor] = self.backbone.detection_features(x)
        stages = self.backbone.return_stages
        feature_list = [features[stage] for stage in stages]

        # Align multi-scale backbone features to a common resolution (LW-DETR expects same spatial size here)
        if len(feature_list) > 1:
            ref_idx = len(feature_list) // 2
            ref_h, ref_w = feature_list[ref_idx].shape[-2:]
            feature_list = [
                (
                    F.interpolate(feat, size=(ref_h, ref_w), mode="nearest")
                    if feat.shape[-2:] != (ref_h, ref_w)
                    else feat
                )
                for feat in feature_list
            ]

        proj_feats = self.projector(feature_list)

        mask_list: Optional[list[torch.Tensor]] = None
        if masks is not None:
            mask_list = []
            for feat in proj_feats:
                mask = F.interpolate(masks[None].float(), size=feat.shape[-2:], mode="nearest").to(torch.bool)[0]
                mask_list.append(mask)

        decoder_feats: list[torch.Tensor] = []
        for idx, proj in enumerate(self.decoder_input_proj):
            decoder_feats.append(proj(proj_feats[idx]))

        spatial_shapes: list[list[int]] = []
        for feat in decoder_feats:
            spatial_shapes.append([feat.shape[-2], feat.shape[-1]])

        level_start_index = [0]
        for shape in spatial_shapes[:-1]:
            level_start_index.append(level_start_index[-1] + shape[0] * shape[1])

        if mask_list is not None:
            valid_ratios = torch.stack([self._get_valid_ratio(m) for m in mask_list], dim=1)
        else:
            B = x.size(0)
            num_levels = len(decoder_feats)
            valid_ratios = torch.ones((B, num_levels, 2), device=proj_feats[0].device, dtype=torch.float32)

        # Two-stage: generate encoder proposals and select top-k per group
        enc_cls_logits: Optional[torch.Tensor] = None
        enc_box_output: Optional[torch.Tensor] = None
        enc_reference_points: Optional[torch.Tensor] = None

        if self.two_stage is True:
            # Flatten decoder features to create memory
            memory_list = []
            mask_flatten_list = []
            for idx, feat in enumerate(decoder_feats):
                memory_list.append(feat.flatten(2).permute(0, 2, 1))
                if mask_list is not None:
                    mask_flatten_list.append(mask_list[idx].flatten(1))

            memory = torch.concat(memory_list, dim=1)
            memory_padding_mask = torch.concat(mask_flatten_list, dim=1) if len(mask_flatten_list) > 0 else None

            # Generate encoder output proposals
            output_memory, output_proposals = self.gen_encoder_output_proposals(
                memory, memory_padding_mask, spatial_shapes
            )

            # Handle case where spatial positions < num_queries
            num_proposals = output_memory.size(1)
            topk = min(self.num_queries, num_proposals)

            if self.training is True:
                # Apply per-group encoder heads and select top-k per group
                enc_cls_list: list[torch.Tensor] = []
                enc_box_list: list[torch.Tensor] = []
                refpoint_list: list[torch.Tensor] = []

                for enc_output, enc_output_norm, enc_out_class, enc_out_bbox in zip(
                    self.enc_output, self.enc_output_norm, self.enc_out_class_embed, self.enc_out_bbox_embed
                ):
                    # Apply per-group encoder output projection
                    output_memory_g = enc_output_norm(enc_output(output_memory))

                    # Compute class and bbox predictions using bbox_reparam formula
                    enc_cls_g = enc_out_class(output_memory_g)
                    enc_box_delta_g = enc_out_bbox(output_memory_g)
                    # bbox_reparam: delta * proposal_wh + proposal_xy for center, exp(delta) * proposal_wh for size
                    enc_box_cxcy_g = enc_box_delta_g[..., :2] * output_proposals[..., 2:] + output_proposals[..., :2]
                    enc_box_wh_g = enc_box_delta_g[..., 2:].exp() * output_proposals[..., 2:]
                    enc_box_g = torch.concat([enc_box_cxcy_g, enc_box_wh_g], dim=-1)

                    # Select top-k proposals for this group
                    _, topk_indices_g = torch.topk(enc_cls_g.max(dim=-1)[0], topk, dim=1)

                    # Gather selected proposals for this group
                    topk_cls_g = torch.gather(
                        enc_cls_g, 1, topk_indices_g.unsqueeze(-1).expand(-1, -1, enc_cls_g.size(-1))
                    )
                    topk_box_g = torch.gather(enc_box_g, 1, topk_indices_g.unsqueeze(-1).expand(-1, -1, 4))

                    enc_cls_list.append(topk_cls_g)
                    enc_box_list.append(topk_box_g)
                    # Reference points (boxes are already in normalized space with bbox_reparam)
                    refpoint_list.append(topk_box_g.detach())

                # Concatenate all groups
                enc_cls_logits = torch.concat(enc_cls_list, dim=1)
                enc_box_output = torch.concat(enc_box_list, dim=1)
                enc_reference_points = torch.concat(refpoint_list, dim=1)
            else:
                # In inference, use only first group (TorchScript-friendly, avoids extra compute)
                enc_output = self.enc_output[0]
                enc_output_norm = self.enc_output_norm[0]
                enc_out_class = self.enc_out_class_embed[0]
                enc_out_bbox = self.enc_out_bbox_embed[0]

                output_memory_g = enc_output_norm(enc_output(output_memory))
                enc_cls_g = enc_out_class(output_memory_g)
                enc_box_delta_g = enc_out_bbox(output_memory_g)
                enc_box_cxcy_g = enc_box_delta_g[..., :2] * output_proposals[..., 2:] + output_proposals[..., :2]
                enc_box_wh_g = enc_box_delta_g[..., 2:].exp() * output_proposals[..., 2:]
                enc_box_g = torch.concat([enc_box_cxcy_g, enc_box_wh_g], dim=-1)

                _, topk_indices_g = torch.topk(enc_cls_g.max(dim=-1)[0], topk, dim=1)
                enc_cls_logits = torch.gather(
                    enc_cls_g, 1, topk_indices_g.unsqueeze(-1).expand(-1, -1, enc_cls_g.size(-1))
                )
                enc_box_output = torch.gather(enc_box_g, 1, topk_indices_g.unsqueeze(-1).expand(-1, -1, 4))
                enc_reference_points = enc_box_output.detach()

        out_bboxes, out_logits = self.decoder(
            decoder_feats,
            spatial_shapes,
            level_start_index,
            valid_ratios,
            padding_mask=mask_list,
            enc_reference_points=enc_reference_points,
            return_intermediates=self.training is True,
        )

        return (enc_cls_logits, enc_box_output, out_bboxes, out_logits)

    def forward(
        self,
        x: torch.Tensor,
        targets: Optional[list[dict[str, torch.Tensor]]] = None,
        masks: Optional[torch.Tensor] = None,
        image_sizes: Optional[list[tuple[int, int]]] = None,
    ) -> tuple[list[dict[str, torch.Tensor]], dict[str, torch.Tensor]]:
        self._input_check(targets)
        image_sizes_tensor = self._to_img_list(x, image_sizes).image_sizes

        enc_cls_logits, enc_box_output, out_bboxes, out_logits = self.forward_net(x, masks)

        losses: dict[str, torch.Tensor] = {}
        detections: list[dict[str, torch.Tensor]] = []
        if self.training is True:
            assert targets is not None, "targets should not be none when in training mode"

            # Compute all losses (encoder + decoder) in compute_loss
            losses = self.compute_loss(
                targets,
                out_logits,
                out_bboxes,
                image_sizes_tensor,
                enc_cls_logits=enc_cls_logits,
                enc_box_output=enc_box_output,
            )
        else:
            detections = self.postprocess_detections(out_logits, out_bboxes, image_sizes_tensor)

        return (detections, losses)


registry.register_model_config(
    "lw_detr",
    LW_DETR,
    config={"projector_scale_factors": [1.0], "decoder_cross_heads": 16},
)
registry.register_model_config(
    "lw_detr_2stg",
    LW_DETR,
    config={"projector_scale_factors": [1.0], "decoder_cross_heads": 16, "two_stage": True},
)
registry.register_model_config(
    "lw_detr_l",
    LW_DETR,
    config={
        "projector_dim": 384,
        "projector_scale_factors": [2.0, 0.5],
        "decoder_dim": 384,
        "decoder_self_heads": 12,
        "decoder_cross_heads": 24,
        "decoder_points": 4,
    },
)
registry.register_model_config(
    "lw_detr_l_2stg",
    LW_DETR,
    config={
        "projector_dim": 384,
        "projector_scale_factors": [2.0, 0.5],
        "decoder_dim": 384,
        "decoder_self_heads": 12,
        "decoder_cross_heads": 24,
        "decoder_points": 4,
        "two_stage": True,
    },
)



"""
Projected Gradient Descent (PGD)

Paper "Towards Deep Learning Models Resistant to Adversarial Attacks", https://arxiv.org/abs/1706.06083
"""

# Reference license: MIT

from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from birder.adversarial.base import AttackResult
from birder.adversarial.base import attack_success
from birder.adversarial.base import clamp_normalized
from birder.adversarial.base import pixel_eps_to_normalized
from birder.adversarial.base import predict_labels
from birder.adversarial.base import validate_target
from birder.data.transforms.classification import RGBType


class PGD:
    def __init__(
        self,
        net: nn.Module,
        eps: float,
        steps: int = 10,
        step_size: Optional[float] = None,
        random_start: bool = False,
        *,
        rgb_stats: RGBType,
    ) -> None:
        if steps <= 0:
            raise ValueError("steps must be a positive integer")

        self.net = net.eval()
        self.eps = eps
        self.steps = steps
        if step_size is not None:
            self.step_size = step_size
        else:
            self.step_size = eps / steps

        self.random_start = random_start
        self.rgb_stats = rgb_stats

        if self.step_size <= 0:
            raise ValueError("step_size must be positive")

    def __call__(self, input_tensor: torch.Tensor, target: Optional[torch.Tensor]) -> AttackResult:
        inputs = input_tensor.detach()
        with torch.no_grad():
            logits = self.net(inputs)

        targeted = target is not None
        if targeted:
            target = validate_target(target, inputs.shape[0], logits.shape[1], inputs.device)
        else:
            target = predict_labels(logits)

        eps_norm = pixel_eps_to_normalized(self.eps, self.rgb_stats, device=inputs.device, dtype=inputs.dtype)
        step_norm = pixel_eps_to_normalized(self.step_size, self.rgb_stats, device=inputs.device, dtype=inputs.dtype)

        # Targeted steps descend toward target, untargeted ascend away from original
        if targeted is True:
            direction = -1.0
        else:
            direction = 1.0

        adv_inputs = inputs.clone()
        if self.random_start is True:
            # Random start inside the epsilon ball
            adv_inputs = adv_inputs + torch.empty_like(adv_inputs).uniform_(-1.0, 1.0) * eps_norm
            adv_inputs = clamp_normalized(adv_inputs, self.rgb_stats)

        for _ in range(self.steps):
            adv_inputs.requires_grad_(True)
            adv_logits = self.net(adv_inputs)
            loss = F.cross_entropy(adv_logits, target)
            (grad,) = torch.autograd.grad(loss, adv_inputs, retain_graph=False, create_graph=False)
            adv_inputs = adv_inputs.detach() + direction * step_norm * grad.sign()

            # Project back into the epsilon ball around the original input.
            delta = torch.clamp(adv_inputs - inputs, min=-eps_norm, max=eps_norm)
            adv_inputs = clamp_normalized(inputs + delta, self.rgb_stats)

        with torch.no_grad():
            adv_logits = self.net(adv_inputs)

        success = attack_success(
            logits.detach(),
            adv_logits,
            targeted,
            target=target if targeted else None,
        )

        return AttackResult(
            adv_inputs=adv_inputs,
            adv_logits=adv_logits,
            perturbation=adv_inputs - inputs,
            logits=logits.detach(),
            success=success,
        )
